"""`StepImplementer` for the `create-container-image` step using Maven JKube Plugin.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key             | Required? | Default | Description
------------------------------|-----------|---------|-----------
`pom-file`                    | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                  | No        | `True`  | Disables TLS Verification if set to False
`maven-profiles`              | No        | `[]`    | List of maven profiles to use.
`maven-no-transfer-progress`  | No        | `True`  | `True` to suppress the transfer progress of packages maven downloads.
                                                     `False` to have the transfer progress printed.\
                                                    See https://maven.apache.org/ref/current/maven-embedder/cli.html
`maven-additional-arguments`  | No        | `['-Dmaven.test.skip=true', '-Dmaven.install.skip=true']` \
                                                    | List of additional arguments to use. \
                                                      Skipping tests by default because assuming \
                                                      a previous step already ran them. \
                                                      Skipping install backs assuming this is \
                                                      running in an ephermal environment where \
                                                      that would be a waist of time, and also \
                                                      that a previous step ran `package` \
                                                      and `push-artifacts` steps.
`maven-servers`               | No        |         | Dictionary of dictionaries of id, username, password
`maven-repositories`          | No        |         | Dictionary of dictionaries of id, url, snapshots, releases
`maven-mirrors`               | No        |         | Dictionary of dictionaries of id, url, mirror_of
`container-image-version`     | Yes       |         | Version to use when building the container image
`organization`                | Yes       |         | Used in built container image tag
`application_name`            | Yes       |         | Used in built container image tag
`service_name`                | Yes       |         | Used in built container image tag
`container-registries`        | No        |         | Hash of container registries to authenticate with.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key                     | Description
----------------------------------------|------------
`container-image-registry-uri`          | Registry URI poriton of the container image tag of the built container image.
`container-image-registry-organization` | Organization portion of the container image tag of the built container image.
`container-image-repository`            | Repository portion of the container image tag of the built container image.
`container-image-name`                  | Another way to reference the repository portion of the container image tag \
                                          of the built container image.
`container-image-version`               | Version portion of the container image tag of the built container image.
`container-image-tag`                   | Full container image tag of the built container, including the registry URI. <br/> \
                                          Takes the form of: \
                                          `container-image-registry-organization/container-image-repository:container-image-version`
`container-image-short-tag`             | Short container image tag of the built container image,  excluding the registry URI. <br/> \
                                          Takes the form of: \
                                          `container-image-registry-uri/container-image-registry-organization/container-image-repository:container-image-version`

"""# pylint: disable=line-too-long

from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementers.shared.maven_generic import \
    MavenGeneric
from ploigos_step_runner.utils.containers import \
    determine_container_image_build_tag_info

DEFAULT_CONFIG = {
    'maven-additional-arguments': [
        '-Dmaven.install.skip=true',
        '-Dmaven.test.skip=true'
    ]
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'pom-file'
]

class MavenJKubeK8sBuild(MavenGeneric):
    """`StepImplementer` for the `create-container-image` step using Maven JKube Plugin.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment,
            maven_phases_and_goals=['k8s:build']
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return {**MavenGeneric.step_implementer_config_defaults(), **DEFAULT_CONFIG}

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self): # pylint: disable=too-many-locals
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # create local build tag
        image_registry_organization = self.get_value('organization')
        build_full_tag, build_short_tag, image_registry_uri, image_repository, image_version = \
            determine_container_image_build_tag_info(
                image_version=self.get_value('container-image-version'),
                organization=image_registry_organization,
                application_name=self.get_value('application-name'),
                service_name=self.get_value('service-name')
            )

        # push the artifacts
        mvn_jkube_output_file_path = self.write_working_file('mvn_k8s_build_output.txt')
        try:
            # execute maven step (params come from config)
            print("Build container image with Maven Jkube kubernetes plugin")
            self._run_maven_step(
                mvn_output_file_path=mvn_jkube_output_file_path,
                step_implementer_additional_arguments=[
                    f"-Djkube.generator.name={build_full_tag}"
                ]
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running 'maven k8s:build' to create container image. " \
                f"More details maybe found in 'maven-jkube-output' report artifact: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "create container image.",
                name='maven-jkube-output',
                value=mvn_jkube_output_file_path
            )

        # add artifacts
        step_result.add_artifact(
            name='container-image-registry-uri',
            value=image_registry_uri,
            description='Registry URI poriton of the container image tag' \
                ' of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-registry-organization',
            value=image_registry_organization,
            description='Organization portion of the container image tag' \
                ' of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-repository',
            value=image_repository,
            description='Repository portion of the container image tag' \
                ' of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-name',
            value=image_repository,
            description='Another way to reference the' \
                ' repository portion of the container image tag of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-version',
            value=image_version,
            description='Version portion of the container image tag of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-tag',
            value=build_full_tag,
            description='Full container image tag of the built container,' \
                ' including the registry URI.'
        )
        step_result.add_artifact(
            name='container-image-short-tag',
            value=build_short_tag,
            description='Short container image tag of the built container image,' \
                ' excluding the registry URI.'
        )

        return step_result



import os
import zipfile
from pathlib import Path
from unittest.mock import patch

import mock
from ploigos_step_runner import StepResult
from ploigos_step_runner.results.workflow_result import WorkflowResult
from ploigos_step_runner.step_implementers.generate_evidence import GenerateEvidence
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


class TestStepImplementerGenerateEvidenceBase(BaseStepImplementerTestCase):

    DEST_URL = 'https://ploigos.com/mock/evidence'
    DEST_URI = 'https://ploigos.com/mock/evidence/test-ORG/test-APP/'\
                'test-SERVICE/test-ORG-test-APP-test-SERVICE-42.0-test-evidence.json'
    GATHER_EVIDENCE_MOCK_DICT = {
                'apiVersion': "automated-governance/v1alpha1",
                'kind': "WorkflowEvidence",
                'workflow': {
                    'test-step': {
                        'attestations': {
                            'test-evidence': {
                                'name': 'test-evidence',
                                'value': 'test-value',
                                'description': 'test-description'
                            },
                            'test-evidence2': {
                                'name': 'test-evidence2',
                                'value': 'test-value2',
                                'description': 'test-description2'
                            }
                        }
                     }
                }
            }

    GATHER_EVIDENCE_MOCK_DICT_WITH_ENVIRONMENT = {
        'apiVersion': "automated-governance/v1alpha1",
        'kind': "WorkflowEvidence",
        'workflow': {
            'test-step': {
                    'attestations': {
                        'test-evidence': {
                            'name': 'test-evidence',
                            'value': 'test-value',
                            'description': 'test-description',
                            'environment': 'foo'
                        },
                        'test-evidence2': {
                            'name': 'test-evidence2',
                            'value': 'test-value2',
                            'description': 'test-description2',
                            'environment': 'foo'
                        }
                    }
            }
        }
    }

    GATHER_EVIDENCE_MOCK_JSON = \
                            """{
    "apiVersion": "automated-governance/v1alpha1",
    "kind": "WorkflowEvidence",
    "workflow": {
        "test-step": {
            "attestations": {
                "test-evidence": {
                    "name": "test-evidence",
                    "value": "test-value",
                    "description": "test-description"
                },
                "test-evidence2": {
                    "name": "test-evidence2",
                    "value": "test-value2",
                    "description": "test-description2"
                }
            }
        }
    }
}"""

    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path='',
            environment=None
    ):
        return self.create_given_step_implementer(
            step_implementer=GenerateEvidence,
            step_config=step_config,
            step_name='generate_evidence',
            implementer='GenerateEvidence',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            environment=environment
        )

class TestStepImplementerGenerateEvidence_other(TestStepImplementerGenerateEvidenceBase):
    def test_step_implementer_config_defaults(self):
        defaults = GenerateEvidence.step_implementer_config_defaults()
        expected_defaults = {
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = GenerateEvidence._required_config_or_result_keys()
        expected_required_keys = [
            'organization',
            'application-name',
            'service-name',
            'version',
        ]
        self.assertEqual(required_keys, expected_required_keys)

class TestStepImplementerGenerateEvidence_run_step(TestStepImplementerGenerateEvidenceBase):

    @patch.object(
        GenerateEvidence,
        '_GenerateEvidence__gather_evidence',
        return_value = None
    )
    def test__run_step_pass_no_evidence(self, generate_evidence_mock):
        step_config = {
            'organization': 'test-ORG',
            'application-name': 'test-APP',
            'service-name': 'test-SERVICE',
            'version': '42.0-test'
        }
        step_implementer = self.create_step_implementer(
            step_config=step_config,
        )

        #Set mock method to return None
        generate_evidence_mock.return_value = None

        step_result = step_implementer._run_step()
        expected_step_result = StepResult(
            step_name='generate_evidence',
            sub_step_name='GenerateEvidence',
            sub_step_implementer_name='GenerateEvidence'
        )


        expected_step_result.add_artifact(
            name='result-generate-evidence',
            value='No evidence to generate.',
            description='Evidence from all previously run steps.'
        )
        expected_step_result.message = "No evidence generated from previously run steps"

        self.assertEqual(step_result, expected_step_result)

        generate_evidence_mock.assert_called_once()

    @patch.object(
        GenerateEvidence,
        '_GenerateEvidence__gather_evidence',
        return_value = "mock return value"
    )
    def test__run_step_pass_with_evidence(self, generate_evidence_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path
            )

            step_result = step_implementer._run_step()
            expected_step_result = StepResult(
                step_name='generate_evidence',
                sub_step_name='GenerateEvidence',
                sub_step_implementer_name='GenerateEvidence'
            )

            expected_step_result.message = 'Evidence successfully packaged ' \
            'in JSON file but was not uploaded to data store (no '\
            'destination URI specified).'

            expected_step_result.add_artifact(
                name='evidence-path',
                value=os.path.join(parent_work_dir_path, 'generate_evidence',
                "test-ORG-test-APP-test-SERVICE-42.0-test-evidence.json"),
                description='File path of evidence.'
            )

            print(str(step_result) + "\n\n" + str(expected_step_result))

            self.assertEqual(step_result, expected_step_result)

            generate_evidence_mock.assert_called_once()

    @patch.object(
        GenerateEvidence,
        '_GenerateEvidence__gather_evidence',
        return_value = TestStepImplementerGenerateEvidenceBase.GATHER_EVIDENCE_MOCK_DICT
    )
    def test_run_step_write_results_to_json_file(self, gather_evidence_mock):

        expected_evidence = self.GATHER_EVIDENCE_MOCK_JSON

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )


            # run the step
            step_result = step_implementer._run_step()

            json_file_path = step_result.get_artifact_value("evidence-path")

            with open(json_file_path, 'r') as actual_json_file:
                json_file_contents = actual_json_file.read()
                print(json_file_contents)
                self.assertEqual(json_file_contents, expected_evidence)
        gather_evidence_mock.assert_called_once()
    @patch('ploigos_step_runner.step_implementers.generate_evidence.generate_evidence.upload_file')
    @patch.object(
        GenerateEvidence,
        '_GenerateEvidence__gather_evidence',
        return_value = TestStepImplementerGenerateEvidenceBase.GATHER_EVIDENCE_MOCK_DICT
    )
    def test__run_step_upload_to_file(self, gather_evidence_mock, upload_file_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
                'evidence-destination-url': self.DEST_URL
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path
            )

            # mock the upload results
            upload_file_mock.return_value = "mock upload results"

            # run the step
            step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='generate_evidence',
                sub_step_name='GenerateEvidence',
                sub_step_implementer_name='GenerateEvidence'
            )
            expected_step_result.add_artifact(
                name='evidence-upload-results',
                description='Results of uploading the evidence JSON file ' \
                            'to the given destination.',
                value='mock upload results'
                )
            expected_step_result.add_artifact(
                name='evidence-uri',
                description='URI of the uploaded results archive.',
                value=self.DEST_URI
            )
            expected_step_result.add_artifact(
                name='evidence-path',
                value=os.path.join(parent_work_dir_path, 'generate_evidence',
                "test-ORG-test-APP-test-SERVICE-42.0-test-evidence.json"),
                description='File path of evidence.'
                )
            expected_step_result.message = "Evidence successfully packaged in JSON file and uploaded to data store."

            self.assertEqual(step_result, expected_step_result)

            # verify mocks called
            gather_evidence_mock.assert_called_once()
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri=self.DEST_URI,
                username=None,
                password=None
            )
    @patch('ploigos_step_runner.step_implementers.generate_evidence.generate_evidence.upload_file')
    @patch.object(
        GenerateEvidence,
        '_GenerateEvidence__gather_evidence',
        return_value = TestStepImplementerGenerateEvidenceBase.GATHER_EVIDENCE_MOCK_DICT
    )
    def test__upload_to_remote_with_auth(self, gather_evidence_mock, upload_file_mock):
        with TempDirectory() as temp_dir:

            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
                'evidence-destination-url': self.DEST_URL,
                'evidence-destination-username': 'test-user',
                'evidence-destination-password': 'test-pass'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path
            )

            # mock the upload results
            upload_file_mock.return_value = "mock upload results"

            # run the step
            step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='generate_evidence',
                sub_step_name='GenerateEvidence',
                sub_step_implementer_name='GenerateEvidence'
            )
            expected_step_result.add_artifact(
                name='evidence-upload-results',
                description='Results of uploading the evidence JSON file ' \
                            'to the given destination.',
                value='mock upload results'
                )
            expected_step_result.add_artifact(
                name='evidence-uri',
                description='URI of the uploaded results archive.',
                value=self.DEST_URI
            )
            expected_step_result.add_artifact(
                name='evidence-path',
                value=os.path.join(parent_work_dir_path, 'generate_evidence',
                "test-ORG-test-APP-test-SERVICE-42.0-test-evidence.json"),
                description='File path of evidence.'
                )
            expected_step_result.message = "Evidence successfully packaged in JSON file and uploaded to data store."

            self.assertEqual(step_result, expected_step_result)

            # verify mocks called
            gather_evidence_mock.assert_called_once()
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri=self.DEST_URI,
                username='test-user',
                password='test-pass'
            )

    @patch('ploigos_step_runner.step_implementers.generate_evidence.generate_evidence.upload_file')
    @patch.object(
        GenerateEvidence,
        '_GenerateEvidence__gather_evidence',
        return_value = TestStepImplementerGenerateEvidenceBase.GATHER_EVIDENCE_MOCK_DICT
    )
    def test__run_step_pass_upload_to_remote_with_auth_failure(self, gather_evidence_mock, upload_file_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
                'evidence-destination-url': self.DEST_URL,
                'evidence-destination-username': 'test-user',
                'evidence-destination-password': 'test-pass'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path
            )

            # mock the upload results
            upload_file_mock.side_effect = RuntimeError('mock upload error')
            upload_file_mock.return_value = "mock upload results"

            # run the step
            step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='generate_evidence',
                sub_step_name='GenerateEvidence',
                sub_step_implementer_name='GenerateEvidence'
            )

            expected_step_result.success = False
            expected_step_result.message = 'mock upload error'

            self.assertEqual(step_result, expected_step_result)

            # verify mocks called
            gather_evidence_mock.assert_called_once()
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri=self.DEST_URI,
                username='test-user',
                password='test-pass'
            )

class TestStepImplementerGenerateEvidence_gather_evidence(TestStepImplementerGenerateEvidenceBase):


    def __setup_evidence(self, parent_work_dir_path, evidence=True, environment=None):
        step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test',
            }

        step_result = StepResult(
            step_name='test-step',
            sub_step_name='test-sub-step',
            sub_step_implementer_name='test-sub-step-implementer',
            environment=environment
        )

        step_result.add_evidence(
            name='test-evidence',
            value="test-value",
            description="test-description"
        )
        step_result.add_evidence(
            name='test-evidence2',
            value="test-value2",
            description="test-description2"
        )

        workflow_result = WorkflowResult()
        workflow_result.add_step_result(step_result)


        step_implementer = self.create_step_implementer(
            step_config=step_config,
            parent_work_dir_path=parent_work_dir_path,
            workflow_result=workflow_result,
        )

        return step_implementer
    def test__gather_evidence(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer =  self.__setup_evidence(parent_work_dir_path)

            gathered_evidence = step_implementer._GenerateEvidence__gather_evidence()

            expected_gathered_evidence = TestStepImplementerGenerateEvidenceBase.GATHER_EVIDENCE_MOCK_DICT

            self.assertEqual(gathered_evidence, expected_gathered_evidence)

    def test___gather_evidence_no_results(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE',
                'version': '42.0-test'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            gathered_evidence = step_implementer._GenerateEvidence__gather_evidence()

            self.assertIsNone(gathered_evidence)

    def test__gather_evidence_with_environment(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer =  self.__setup_evidence(parent_work_dir_path, environment='foo')

            gathered_evidence = step_implementer._GenerateEvidence__gather_evidence()

            expected_gathered_evidence = TestStepImplementerGenerateEvidenceBase.GATHER_EVIDENCE_MOCK_DICT_WITH_ENVIRONMENT

            self.assertEqual(gathered_evidence, expected_gathered_evidence)



"""
Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key            | Required? | Default          | Description
-----------------------------|-----------|------------------|-----------
`package-file`               | No        | `'package.json'` | package.json file for reference app.
`npm-args`                   | Yes       | `None`           | Arguments to pass to the npm command.
"""


import os
from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementer import StepImplementer
from ploigos_step_runner.utils.npm import run_npm

DEFAULT_CONFIG = {
    'package-file': 'package.json'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'package-file',
    'npm-args'
]

class NpmGeneric(StepImplementer):
    """Abstract parent class for StepImplementers that use npm.
    """

    def __init__( # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None,
        npm_args=None
    ):
        self.__npm_args = npm_args

        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * given 'package-file' exists

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        # if package-file has value verify file exists
        # If it doesn't have value and is required function will have already failed
        package_file = self.get_value('package-file')
        if package_file is not None:
            assert os.path.exists(package_file), \
                f'Given npm package file (package-file) does not exist: {package_file}'

    @property
    def npm_args(self):
        """Property for getting the npm phases and goals to execute which can either come
        from field set on this class via constructor, intended for use by sub classes that want
        to hard code the phases and goals for convenience, or comes from config value
        `npm-phases-and-goals` set by the user.

        Returns
        -------
        str
            Maven phases and/or goals to execute.
        """
        npm_args = None
        if self.__npm_args:
            npm_args = self.__npm_args
        else:
            npm_args = self.get_value('npm-args')

        return npm_args

    def _run_npm_step(
        self,
        npm_output_file_path
    ):
        """Runs npm using the configuration given to this step runner.

        Parameters
        ----------
        npm_output_file_path : str
            Path to file containing the npm stdout and stderr output.
        step_implementer_additional_arguments : []
            Additional arguments hard coded by the step implementer.

        Raises
        ------
        StepRunnerException
            If npm returns a none 0 exit code.
        """

        npm_args = self.npm_args

        run_npm(
            npm_output_file_path=npm_output_file_path,
            npm_args=npm_args
        )

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # package the artifacts
        npm_output_file_path = self.write_working_file('npm_output.txt')
        try:
            # execute npm step (params come from config)
            self._run_npm_step(
                npm_output_file_path=npm_output_file_path
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running npm. " \
                f"More details maybe found in 'npm-output' report artifact: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from npm.",
                name='npm-output',
                value=npm_output_file_path
            )

        return step_result



import os
from contextlib import redirect_stdout
from io import StringIO
from unittest.mock import patch

from ploigos_step_runner import StepResult, WorkflowResult
from ploigos_step_runner.config import Config
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.step_implementer import StepImplementer
from ploigos_step_runner.step_runner import StepRunner
from testfixtures import TempDirectory

from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.sample_step_implementers import (
    FailStepImplementer, FooStepImplementer,
    RequiredStepConfigMultipleOptionsStepImplementer,
    WriteConfigAsResultsStepImplementer)


class TestStepImplementer(BaseStepImplementerTestCase):
    def _run_step_implementer_test(
        self,
        config,
        step,
        expected_step_results,
        test_dir,
        environment=None
    ):
        working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
        factory = StepRunner(
            config=config,
            results_file_name='step-runner-results.yml',
            work_dir_path=working_dir_path
        )
        factory.run_step(
            step_name=step,
            environment=environment
        )

        pickle = f'{working_dir_path}/step-runner-results.pkl'
        workflow_result = WorkflowResult.load_from_pickle_file(pickle)

        step_result = workflow_result.get_step_result(
            step_name=step
        )
        self.assertEqual(step_result, expected_step_results)

    def _setup_get_value_with_env_test(self, test_dir, environment):
        step_config = {
            'test': 'hello world',
            'old-param-name': 'foo42',
            'new-param-name': 'bar42'
        }

        parent_work_dir_path = os.path.join(test_dir.path, 'working')

        workflow_result = WorkflowResult()

        step_result_no_evn = StepResult(
            step_name='foo',
            sub_step_name='Mock',
            sub_step_implementer_name='Mock'
        )
        step_result_no_evn.add_artifact(
            name='container-image-tag',
            value='localhost/mock:0.42.0-weird'
        )
        workflow_result.add_step_result(step_result=step_result_no_evn)

        step_result_deploy_test = StepResult(
            step_name='deploy',
            sub_step_name='ArgoCD',
            sub_step_implementer_name='ArgoCD',
            environment='TEST'
        )
        step_result_deploy_test.add_artifact(
            name='deployed-host-urls',
            value='https://awesome-app.test.ploigos.xyz'
        )
        step_result_deploy_test.add_artifact(
            name='container-image-tag',
            value='localhost/test/mock:0.42.0-weird'
        )
        workflow_result.add_step_result(step_result=step_result_deploy_test)

        step_result_deploy_prod = StepResult(
            step_name='deploy',
            sub_step_name='ArgoCD',
            sub_step_implementer_name='ArgoCD',
            environment='PROD'
        )
        step_result_deploy_prod.add_artifact(
            name='deployed-host-urls',
            value='https://awesome-app.prod.ploigos.xyz'
        )
        step_result_deploy_test.add_artifact(
            name='container-image-tag',
            value='localhost/prod/mock:0.42.0-weird'
        )
        workflow_result.add_step_result(step_result=step_result_deploy_prod)
        pickle_filename = os.path.join(parent_work_dir_path, 'step-runner-results.pkl')
        workflow_result.write_to_pickle_file(pickle_filename=pickle_filename)

        return self.create_given_step_implementer(
            step_implementer=FooStepImplementer,
            step_config=step_config,
            step_name='foo',
            implementer='FooStepImplementer',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            environment=environment
        )

@patch.object(StepImplementer, '_StepImplementer__add_additional_artifacts_to_step_result')
class TestStepImplementer_run_step(TestStepImplementer):
    def test_one_step_writes_to_empty_results_file(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config1 = {
            'step-runner-config': {
                'write-config-as-results': {
                    'implementer':
                        'tests.helpers.sample_step_implementers.'
                        'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'config-1': "config-1",
                        'config-overwrite-me': 'config-1',
                        'required-config-key': 'required'
                    }
                }
            }
        }
        config1_expected_step_results = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer'
        )
        config1_expected_step_results.success=True
        config1_expected_step_results.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results.add_artifact(
            name='required-config-key',
            value='required'
        )

        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results,
                test_dir
            )

        mock_add_additional_artifacts_to_step_result.assert_called_once()

    def test_one_step_existing_results_file_bad_pickle(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config = {
            'step-runner-config': {
                'write-config-as-results': {
                    'implementer': 'tests.helpers.sample_step_implementers.'
                                   'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'config-1': "config-1",
                        'config-overwrite-me': 'config-1',
                        'required-config-key': 'required'
                    }
                }
            }
        }

        with TempDirectory() as test_dir:
            results_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            results_file_path = os.path.join(results_dir_path, 'step-runner-results.pkl')
            test_dir.write(results_file_path, b'''{}bad[yaml}''')

            with self.assertRaisesRegex(
                    StepRunnerException,
                    r'error loading .*'):
                self._run_step_implementer_test(
                    config,
                    'write-config-as-results',
                    None,
                    test_dir
                )

        mock_add_additional_artifacts_to_step_result.assert_not_called()

    def test_boolean_false_config_variable(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config = {
            'step-runner-config': {
                'write-config-as-results': {
                    'implementer': 'tests.helpers.sample_step_implementers.'
                                   'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'required-config-key': False
                    }
                }
            }
        }

        config_expected_step_results = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer'
        )
        config_expected_step_results.success=True
        config_expected_step_results.add_artifact(
            name='required-config-key',
            value=False
        )

        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config,
                'write-config-as-results',
                config_expected_step_results,
                test_dir
            )

        mock_add_additional_artifacts_to_step_result.assert_called_once()

    def test_one_step_existing_results_file_empty(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config = {
            'step-runner-config': {
                'write-config-as-results': {
                    'implementer': 'tests.helpers.sample_step_implementers.'
                                   'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'config-1': "config-1",
                        'config-overwrite-me': 'config-1',
                        'required-config-key': 'required'
                    }
                }
            }
        }

        config_expected_step_results = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer'
        )
        config_expected_step_results.success = True
        config_expected_step_results.add_artifact(
            name='config-1',
            value='config-1'
        )
        config_expected_step_results.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config_expected_step_results.add_artifact(
            name='required-config-key',
            value='required'
        )

        with TempDirectory() as test_dir:
            results_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            results_file_path = os.path.join(results_dir_path, 'step-runner-results.pkl')
            test_dir.write(results_file_path, b'''''')
            self._run_step_implementer_test(
                config,
                'write-config-as-results',
                config_expected_step_results,
                test_dir
            )

        mock_add_additional_artifacts_to_step_result.assert_called_once()

    def test_global_environment_default_config(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config1 = {
            'step-runner-config': {
                'global-environment-defaults': {
                    'SAMPLE-ENV-1': {
                        'sample-config-option-1': 'sample env 1 value'
                    },
                    'SAMPLE-ENV-2': {
                        'sample-config-option-1': 'sample env 2 value'
                    }
                },
                'write-config-as-results': {
                    'implementer':
                        'tests.helpers.sample_step_implementers.'
                        'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'config-1': "config-1",
                        'config-overwrite-me': 'config-1',
                        'required-config-key': 'required'
                    }
                }
            }
        }
        config1_expected_step_results_env_1 = {
            'SAMPLE-ENV-1': {
                'write-config-as-results': {
                    'tests.helpers.sample_step_implementers.'
                    'WriteConfigAsResultsStepImplementer': {
                        'sub-step-implementer-name':
                            'tests.helpers.sample_step_implementers.'
                            'WriteConfigAsResultsStepImplementer',
                        'success': True,
                        'message': '',
                        'artifacts': {
                            'environment-name':
                                {'description': '', 'value': 'SAMPLE-ENV-1'},
                            'sample-config-option-1':
                                {'description': '', 'value': 'sample env 1 value'},
                            'config-1':
                                {'description': '', 'value': 'config-1'},
                            'config-overwrite-me':
                                {'description': '', 'value': 'config-1'},
                            'required-config-key':
                                {'description': '', 'value': 'required'},
                        }
                    }
                }
            }
        }
        config1_expected_step_results_env_2 = {
            'SAMPLE-ENV-2': {
                'write-config-as-results': {
                    'tests.helpers.sample_step_implementers.'
                    'WriteConfigAsResultsStepImplementer': {
                        'sub-step-implementer-name':
                            'tests.helpers.sample_step_implementers.'
                            'WriteConfigAsResultsStepImplementer',
                        'success': True,
                        'message': '',
                        'artifacts': {
                            'environment-name':
                                {'description': '', 'value': 'SAMPLE-ENV-2'},
                            'config-1':
                                {'description': '', 'value': 'config-1'},
                            'sample-config-option-1':
                                {'description': '', 'value': 'sample env 2 value'},
                            'config-overwrite-me':
                                {'description': '', 'value': 'config-1'},
                            'required-config-key':
                                {'description': '', 'value': 'required'}
                        }
                    }
                }
            }
        }

        config1_expected_step_results_env_1 = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            environment='SAMPLE-ENV-1'
        )
        config1_expected_step_results_env_1.success=True
        config1_expected_step_results_env_1.add_artifact(
            name='environment-name',
            value='SAMPLE-ENV-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='sample-config-option-1',
            value='sample env 1 value'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='required-config-key',
            value='required'
        )

        config1_expected_step_results_env_2 = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            environment='SAMPLE-ENV-2'
        )
        config1_expected_step_results_env_2.success=True
        config1_expected_step_results_env_2.add_artifact(
            name='environment-name',
            value='SAMPLE-ENV-2'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='sample-config-option-1',
            value='sample env 2 value'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='required-config-key',
            value='required'
        )

        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results_env_1,
                test_dir,
                'SAMPLE-ENV-1'
            )
        mock_add_additional_artifacts_to_step_result.assert_called_once()

        mock_add_additional_artifacts_to_step_result.reset_mock()
        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results_env_2,
                test_dir,
                'SAMPLE-ENV-2'
            )
        mock_add_additional_artifacts_to_step_result.assert_called_once()

    def test_step_environment_config(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config1 = {
            'step-runner-config': {
                'write-config-as-results': {
                    'implementer': 'tests.helpers.sample_step_implementers.'
                                   'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'config-1': "config-1",
                        'config-overwrite-me': 'config-1',
                        'required-config-key': 'required'
                    },
                    'environment-config': {
                        'SAMPLE-ENV-1': {
                            'sample-config-option-1': 'step env config - env 1 value'
                        },
                        'SAMPLE-ENV-2': {
                            'sample-config-option-1': 'step env config - env 2 value'
                        }
                    }
                }
            }
        }

        config1_expected_step_results_env_1 = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            environment='SAMPLE-ENV-1'
        )
        config1_expected_step_results_env_1.success=True
        config1_expected_step_results_env_1.add_artifact(
            name='sample-config-option-1',
            value='step env config - env 1 value'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='required-config-key',
            value='required'
        )

        config1_expected_step_results_env_2 = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            environment='SAMPLE-ENV-2'
        )
        config1_expected_step_results_env_2.success=True
        config1_expected_step_results_env_2.add_artifact(
            name='sample-config-option-1',
            value='step env config - env 2 value'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='required-config-key',
            value='required'
        )

        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results_env_1,
                test_dir,
                'SAMPLE-ENV-1'
            )
        mock_add_additional_artifacts_to_step_result.assert_called_once()

        mock_add_additional_artifacts_to_step_result.reset_mock()
        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results_env_2,
                test_dir,
                'SAMPLE-ENV-2'
            )
        mock_add_additional_artifacts_to_step_result.assert_called_once()


    def test_global_environment_default_and_step_environment_config(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config1 = {
            'step-runner-config': {
                'global-environment-defaults': {
                    'SAMPLE-ENV-1': {
                        'sample-config-option-1': 'global env config - env 1 value - 1',
                        'sample-config-option-2': 'global env config - env 1 value - 2'
                    },
                    'SAMPLE-ENV-2': {
                        'sample-config-option-1': 'global env config - env 2 value - 1',
                        'sample-config-option-2': 'global env config - env 2 value - 2'
                    }
                },
                'write-config-as-results': {
                    'implementer': 'tests.helpers.sample_step_implementers.'
                                   'WriteConfigAsResultsStepImplementer',
                    'config': {
                        'config-1': "config-1",
                        'config-overwrite-me': 'config-1',
                        'required-config-key': 'required'
                    },
                    'environment-config': {
                        'SAMPLE-ENV-1': {
                            'sample-config-option-1': 'step env config - env 1 value - 1'
                        },
                        'SAMPLE-ENV-2': {
                            'sample-config-option-1': 'step env config - env 2 value - 1'
                        }
                    },
                }
            }
        }

        config1_expected_step_results_env_1 = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            environment='SAMPLE-ENV-1'
        )
        config1_expected_step_results_env_1.success=True
        config1_expected_step_results_env_1.add_artifact(
            name='environment-name',
            value='SAMPLE-ENV-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='sample-config-option-1',
            value='step env config - env 1 value - 1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='sample-config-option-2',
            value='global env config - env 1 value - 2'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results_env_1.add_artifact(
            name='required-config-key',
            value='required'
        )

        config1_expected_step_results_env_2 = StepResult(
            step_name='write-config-as-results',
            sub_step_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.'
                'WriteConfigAsResultsStepImplementer',
            environment='SAMPLE-ENV-2'
        )
        config1_expected_step_results_env_2.success=True
        config1_expected_step_results_env_2.add_artifact(
            name='environment-name',
            value='SAMPLE-ENV-2'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='sample-config-option-1',
            value='step env config - env 2 value - 1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='sample-config-option-2',
            value='global env config - env 2 value - 2'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='config-1',
            value='config-1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='config-overwrite-me',
            value='config-1'
        )
        config1_expected_step_results_env_2.add_artifact(
            name='required-config-key',
            value='required'
        )

        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results_env_1,
                test_dir,
                'SAMPLE-ENV-1'
            )
        mock_add_additional_artifacts_to_step_result.assert_called_once()

        mock_add_additional_artifacts_to_step_result.reset_mock()
        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config1,
                'write-config-as-results',
                config1_expected_step_results_env_2,
                test_dir,
                'SAMPLE-ENV-2'
            )
        mock_add_additional_artifacts_to_step_result.assert_called_once()

    def test_missing_required_config_key(
        self,
        mock_add_additional_artifacts_to_step_result
    ):
        config = {
            'step-runner-config': {
                'required-step-config-test': {
                    'implementer':
                        'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer',
                    'config': {}
                }
            }
        }

        expected_results = StepResult(
            step_name='required-step-config-test',
            sub_step_name='tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
        )
        expected_results.success = False
        expected_results.message = "Missing required step configuration" \
            " or previous step result artifact keys: ['required-config-key']"

        with TempDirectory() as test_dir:
            self._run_step_implementer_test(
                config,
                'required-step-config-test',
                expected_results,
                test_dir
            )

        mock_add_additional_artifacts_to_step_result.assert_not_called()

class TestStepImplementer_other(TestStepImplementer):
    def test_empty_constructor_params(self):
        config = Config({
            'step-runner-config': {
                'write-config-as-results': {
                    'implementer': 'tests.helpers.sample_step_implementers.'
                                   'WriteConfigAsResultsStepImplementer'
                }
            }
        })
        step_config = config.get_step_config('write-config-as-results')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.'
            'WriteConfigAsResultsStepImplementer')

        step = WriteConfigAsResultsStepImplementer(
            workflow_result=None,
            parent_work_dir_path='',
            config=sub_step
        )

        self.assertEqual(step.step_environment_config, {})
        self.assertEqual(step.step_config, {})
        self.assertEqual(step.global_config_defaults, {})
        self.assertEqual(step.global_environment_config_defaults, {})

    def test_write_working_file(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {}
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        with TempDirectory() as test_dir:
            working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            step = FooStepImplementer(
                workflow_result=WorkflowResult(),
                parent_work_dir_path=working_dir_path,
                config=sub_step
            )

            step.write_working_file('test-working-file', b'hello world')

            with open(os.path.join(working_dir_path, 'foo', 'test-working-file'), 'r') \
                    as working_file:
                self.assertEqual(working_file.read(), 'hello world')

    def test_write_working_file_touch(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {}
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        with TempDirectory() as test_dir:
            working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            step = FooStepImplementer(
                workflow_result=WorkflowResult(),
                parent_work_dir_path=working_dir_path,
                config=sub_step
            )

            step.write_working_file('foo/test.json')

            working_file_path = os.path.join(working_dir_path, 'foo/foo/test.json')
            self.assertTrue(os.path.exists(working_file_path))

            with open(working_file_path, 'r') as working_file:
                self.assertEqual(working_file.read(), '')

    def test_get_config_value(self):
        step_config = {
            'test': 'hello world'
        }

        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step = self.create_given_step_implementer(
                step_implementer=FooStepImplementer,
                step_config=step_config,
                step_name='foo',
                implementer='FooStepImplementer',
                parent_work_dir_path=parent_work_dir_path
            )

            self.assertEqual(step.get_config_value('test'), 'hello world')
            self.assertIsNone(step.get_config_value('does-not-exist'))

    def test_get_config_value_with_env(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {
                        'test': 'hello world',
                        'env-config-override-key': 'override-me'
                    },
                    'environment-config': {
                        'SAMPLE-ENV-1': {
                            'env-config-override-key': 'step env config - env 1 value - 1'
                        },
                        'SAMPLE-ENV-2': {
                            'env-config-override-key': 'step env config - env 2 value - 1'
                        }
                    },
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        with TempDirectory() as test_dir:
            working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            step = FooStepImplementer(
                workflow_result=WorkflowResult(),
                parent_work_dir_path=working_dir_path,
                config=sub_step,
                environment='SAMPLE-ENV-1'
            )

            self.assertEqual(step.get_config_value('test'), 'hello world')
            self.assertIsNone(step.get_config_value('does-not-exist'))
            self.assertEqual(
                step.get_config_value('env-config-override-key'),
                'step env config - env 1 value - 1')

    def test_has_config_value(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {
                        'test': 'hello world',
                        'env-config-override-key': 'override-me',
                        'username': 'foo',
                        'password': 'bar'
                    },
                    'environment-config': {
                        'SAMPLE-ENV-1': {
                            'env-config-override-key': 'step env config - env 1 value - 1'
                        },
                        'SAMPLE-ENV-2': {
                            'env-config-override-key': 'step env config - env 2 value - 1'
                        }
                    },
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        with TempDirectory() as test_dir:
            working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            step = FooStepImplementer(
                workflow_result=WorkflowResult(),
                parent_work_dir_path=working_dir_path,
                config=sub_step,
                environment='SAMPLE-ENV-1'
            )

            self.assertFalse(step.has_config_value('bar'))
            self.assertFalse(step.has_config_value(['bar']))

            self.assertFalse(step.has_config_value(['username', 'foo'], False))
            self.assertTrue(step.has_config_value(['username', 'foo'], True))

            self.assertTrue(step.has_config_value(['username', 'password'], False))
            self.assertTrue(step.has_config_value(['username', 'password'], True))

    def test_fail(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FailStepImplementer',
                    'config': {}
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FailStepImplementer')

        with TempDirectory() as test_dir:
            working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            step = FailStepImplementer(
                workflow_result=WorkflowResult(),
                parent_work_dir_path=working_dir_path,
                config=sub_step
            )
            result = step.run_step()

        self.assertEqual(False, result.success)

    def test_create_working_dir_sub_dir(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {}
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        with TempDirectory() as test_dir:
            working_dir_path = os.path.join(test_dir.path, 'step-runner-working')
            step = FooStepImplementer(
                workflow_result=WorkflowResult(),
                parent_work_dir_path=working_dir_path,
                config=sub_step
            )

            sub = step.create_working_dir_sub_dir('folder1')
            self.assertEqual(sub, f'{working_dir_path}/foo/folder1')
            self.assertEqual(True, os.path.isdir(sub))

            sub = step.create_working_dir_sub_dir('folder1/folder2/folder3')
            self.assertEqual(sub, f'{working_dir_path}/foo/folder1/folder2/folder3')
            self.assertEqual(True, os.path.isdir(sub))

class TestStepImplementer_get_value(TestStepImplementer):
    def test_get_value_no_env(self):
        step_config = {
            'test': 'hello world'
        }

        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            artifact_config = {
                'fake-previous-step-artifact': {
                    'description': '',
                    'value': 'world hello'
                },
            }

            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step = self.create_given_step_implementer(
                step_implementer=FooStepImplementer,
                step_config=step_config,
                step_name='foo',
                implementer='FooStepImplementer',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            self.assertEqual(step.get_value('test'), 'hello world')
            self.assertEqual(step.get_value('fake-previous-step-artifact'), 'world hello')
            self.assertIsNone(step.get_value('does-not-exist'))

    def test_get_value_from_artifact_without_specifing_environment_where_last_artifact_was_from_environment_and_artifact_never_set_without_environment(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment=None)
            self.assertEqual(
                step.get_value('deployed-host-urls'),
                'https://awesome-app.prod.ploigos.xyz'
            )

    # NOTE: maybe this should actually return the first instance where the environment was not set?
    def test_get_value_from_artifact_without_specifing_environment_where_last_artifact_was_from_environment_and_artifact_was_set_without_environment(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment=None)
            self.assertEqual(
                step.get_value('container-image-tag'),
                'localhost/prod/mock:0.42.0-weird'
            )

    def test_get_value_from_step_run_in_environment_that_has_result_value_for_that_environment_1(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='TEST')
            self.assertEqual(
                step.get_value('deployed-host-urls'),
                'https://awesome-app.test.ploigos.xyz'
            )

    def test_get_value_from_step_run_in_environment_that_has_result_value_for_that_environment_2(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='PROD')
            self.assertEqual(
                step.get_value('deployed-host-urls'),
                'https://awesome-app.prod.ploigos.xyz'
            )

    def test_get_value_from_step_run_in_environment_that_nodes_not_have_result_value_for_that_environment(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='RANDOM')
            self.assertEqual(
                step.get_value('deployed-host-urls'),
                'https://awesome-app.prod.ploigos.xyz'
            )

    def test_get_value_multiple_keys_1(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='RANDOM')
            self.assertEqual(
                step.get_value(['old-param-name', 'new-param-name']),
                'foo42'
            )

    def test_get_value_multiple_keys_2(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='RANDOM')
            self.assertEqual(
                step.get_value(['new-param-name', 'old-param-name']),
                'bar42'
            )

    def test_get_value_multiple_keys_3(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='RANDOM')
            self.assertEqual(
                step.get_value(['does-not-exist', 'old-param-name']),
                'foo42'
            )

    def test_get_value_multiple_keys_4(self):
        with TempDirectory() as test_dir:
            step = self._setup_get_value_with_env_test(test_dir=test_dir, environment='RANDOM')
            self.assertEqual(
                step.get_value(['old-param-name', 'does-not-exist']),
                'foo42'
            )

class TestStepImplementer_validate_required_config_or_previous_step_result_artifact_keys(TestStepImplementer):
    def test__validate_required_config_or_previous_step_result_artifact_keys_mutliple_keys_missing_config(self):
        with TempDirectory() as test_dir:
            step_config = {}

            step = self.create_given_step_implementer(
                step_implementer=RequiredStepConfigMultipleOptionsStepImplementer,
                step_config=step_config,
                step_name='foo',
                implementer='RequiredStepConfigMultipleOptionsStepImplementer',
                parent_work_dir_path=test_dir.path
            )

            with self.assertRaisesRegex(
                AssertionError,
                r"Missing required step configuration or previous step result artifact keys: "
                r"\[\['new-required-key', 'old-required-key'\]\]"
            ):
                step._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_mutliple_keys_first_option(self):
        with TempDirectory() as test_dir:
            step_config = {
                'new-required-key': 'new-value-42'
            }

            step = self.create_given_step_implementer(
                step_implementer=RequiredStepConfigMultipleOptionsStepImplementer,
                step_config=step_config,
                step_name='foo',
                implementer='RequiredStepConfigMultipleOptionsStepImplementer',
                parent_work_dir_path=test_dir.path
            )

            step._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_mutliple_keys_second_option(self):
        with TempDirectory() as test_dir:
            step_config = {
                'old-required-key': 'new-value-42'
            }

            step = self.create_given_step_implementer(
                step_implementer=RequiredStepConfigMultipleOptionsStepImplementer,
                step_config=step_config,
                step_name='foo',
                implementer='RequiredStepConfigMultipleOptionsStepImplementer',
                parent_work_dir_path=test_dir.path
            )

            step._validate_required_config_or_previous_step_result_artifact_keys()

class TestStepImplementer___print_data_string(TestStepImplementer):
    def test___print_data_string(self):
        stdout_buff = StringIO()
        with redirect_stdout(stdout_buff):
            StepImplementer._StepImplementer__print_data(
                title="Test Title",
                data="test string data"
            )

        stdout = stdout_buff.getvalue()
        self.assertEqual(
            stdout,
            "        Test Title\n"
            "            test string data\n\n"
        )

    def test___print_data_num(self):
        stdout_buff = StringIO()
        with redirect_stdout(stdout_buff):
            StepImplementer._StepImplementer__print_data(
                title="Test Title",
                data="42"
            )

        stdout = stdout_buff.getvalue()
        self.assertEqual(
            stdout,
            "        Test Title\n"
            "            42\n\n"
        )

    def test___print_data_list(self):
        stdout_buff = StringIO()
        with redirect_stdout(stdout_buff):
            StepImplementer._StepImplementer__print_data(
                title="Test Title",
                data=[
                    'hello',
                    'world',
                    42
                ]
            )

        stdout = stdout_buff.getvalue()
        self.assertEqual(
            stdout,
            '        Test Title\n'
            '            [\n'
            '                "hello",\n'
            '                "world",\n'
            '                42\n'
            '            ]\n\n'
        )

    def test___print_data_dict(self):
        stdout_buff = StringIO()
        with redirect_stdout(stdout_buff):
            StepImplementer._StepImplementer__print_data(
                title="Test Title",
                data={
                    'application-name': 'ref-quarkus-mvn-jenkins-std',
                    'container-registries': {
                        'quay-quay-enterprise.apps.tssc.rht-set.com': {
                            'password': '*************************************',
                            'username': 'ploigos-ref+workflow_reference_quarkus_mvn_jenkins_workflow_standard'
                        },
                        'registry.redhat.io': {
                            'password': '******************************',
                            'username': '6340056|tssc-integration-infra'
                        }
                    },
                    'maven-mirrors': {
                        'internal-mirror': {
                            'id': 'internal-mirror',
                            'mirror-of': '*',
                            'url': 'http://artifactory.apps.tssc.rht-set.com/artifactory/release/'
                        }
                    },
                    'maven-servers': {
                        'internal-mirror': {
                            'id': 'internal-server',
                            'password': '***********',
                            'username': 'tssc'
                        }
                    },
                    'organization': 'ploigos-ref',
                    'service-name': 'fruit'
                }
            )

        stdout = stdout_buff.getvalue()
        self.assertEqual(
            stdout,
            '        Test Title\n'
            '            {\n'
            '                "application-name": "ref-quarkus-mvn-jenkins-std",\n'
            '                "container-registries": {\n'
            '                    "quay-quay-enterprise.apps.tssc.rht-set.com": {\n'
            '                        "password": "*************************************",\n'
            '                        "username": "ploigos-ref+workflow_reference_quarkus_mvn_jenkins_workflow_standard"\n'
            '                    },\n'
            '                    "registry.redhat.io": {\n'
            '                        "password": "******************************",\n'
            '                        "username": "6340056|tssc-integration-infra"\n'
            '                    }\n'
            '                },\n'
            '                "maven-mirrors": {\n'
            '                    "internal-mirror": {\n'
            '                        "id": "internal-mirror",\n'
            '                        "mirror-of": "*",\n'
            '                        "url": "http://artifactory.apps.tssc.rht-set.com/artifactory/release/"\n'
            '                    }\n'
            '                },\n'
            '                "maven-servers": {\n'
            '                    "internal-mirror": {\n'
            '                        "id": "internal-server",\n'
            '                        "password": "***********",\n'
            '                        "username": "tssc"\n'
            '                    }\n'
            '                },\n'
            '                "organization": "ploigos-ref",\n'
            '                "service-name": "fruit"\n'
            '            }\n\n'
        )

class TestStepImplementer___add_additional_artifacts_to_step_result(BaseStepImplementerTestCase):
    def create_step_implementer(
        self,
        step_config={},
        workflow_result=None,
        parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=FooStepImplementer,
            step_config=step_config,
            step_name='mock-step',
            implementer='FooStepImplementer',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_no_additional_artifacts(self):
        # setup test
        step_config = {}
        step_implementer = self.create_step_implementer(
            step_config=step_config,
            parent_work_dir_path='/mock/work-dir'
        )

        # run test
        actual_step_result = StepResult(
            step_name='mock-step',
            sub_step_name='FooStepImplementer',
            sub_step_implementer_name='FooStepImplementer'
        )
        step_implementer._StepImplementer__add_additional_artifacts_to_step_result(
            step_result=actual_step_result
        )

        # verify results
        expected_step_result = StepResult(
            step_name='mock-step',
            sub_step_name='FooStepImplementer',
            sub_step_implementer_name='FooStepImplementer'
        )
        self.assertEqual(actual_step_result, expected_step_result)

    def test_additional_artifacts_list_of_dirs(self):
        # setup test
        step_config = {
            'additional-artifacts': [
                {
                    'name': 'mock-user-reports1',
                    'value': 'target/mock-foo'
                },
                {
                    'name': 'mock-user-reports2',
                    'value': 'target/mock-bar'
                }
            ]
        }
        step_implementer = self.create_step_implementer(
            step_config=step_config,
            parent_work_dir_path='/mock/work-dir'
        )

        # run test
        actual_step_result = StepResult(
            step_name='mock-step',
            sub_step_name='FooStepImplementer',
            sub_step_implementer_name='FooStepImplementer'
        )
        step_implementer._StepImplementer__add_additional_artifacts_to_step_result(
            step_result=actual_step_result
        )

        # verify results
        expected_step_result = StepResult(
            step_name='mock-step',
            sub_step_name='FooStepImplementer',
            sub_step_implementer_name='FooStepImplementer'
        )
        expected_step_result.add_artifact(
            name='mock-user-reports1',
            value='target/mock-foo'
        )
        expected_step_result.add_artifact(
            name='mock-user-reports2',
            value='target/mock-bar'
        )
        self.assertEqual(actual_step_result, expected_step_result)

    def test_additional_artifacts_list_of_strings(self):
        # setup test
        step_config = {
            'additional-artifacts': [
                'target/mock-foo',
                'target/mock-bar'
            ]
        }
        step_implementer = self.create_step_implementer(
            step_config=step_config,
            parent_work_dir_path='/mock/work-dir'
        )

        # run test
        actual_step_result = StepResult(
            step_name='mock-step',
            sub_step_name='FooStepImplementer',
            sub_step_implementer_name='FooStepImplementer'
        )
        step_implementer._StepImplementer__add_additional_artifacts_to_step_result(
            step_result=actual_step_result
        )

        # verify results
        expected_step_result = StepResult(
            step_name='mock-step',
            sub_step_name='FooStepImplementer',
            sub_step_implementer_name='FooStepImplementer'
        )
        expected_step_result.add_artifact(
            name='mock-foo',
            value='target/mock-foo'
        )
        expected_step_result.add_artifact(
            name='mock-bar',
            value='target/mock-bar'
        )
        self.assertEqual(actual_step_result, expected_step_result)



import hashlib
import os
import re
import sys
import sh
from io import IOBase
from tests.helpers.test_utils import Any
from unittest.mock import patch

from ploigos_step_runner.utils.pgp import (detach_sign_with_pgp_key, import_pgp_key, export_pgp_public_key)
from tests.helpers.base_test_case import BaseTestCase
from ploigos_step_runner.exceptions import StepRunnerException

class TestImportPGPKey(BaseTestCase):
    TEST_FAKE_PRIVATE_KEY = '''
        -----BEGIN RSA PRIVATE KEY-----
        MIICXAIBAAKBgQCqGKukO1De7zhZj6+H0qtjTkVxwTCpvKe4eCZ0FPqri0cb2JZfXJ/DgYSF6vUp
        wmJG8wVQZKjeGcjDOL5UlsuusFncCzWBQ7RKNUSesmQRMSGkVb1/3j+skZ6UtW+5u09lHNsj6tQ5
        1s1SPrCBkedbNf0Tp0GbMJDyR4e9T04ZZwIDAQABAoGAFijko56+qGyN8M0RVyaRAXz++xTqHBLh
        3tx4VgMtrQ+WEgCjhoTwo23KMBAuJGSYnRmoBZM3lMfTKevIkAidPExvYCdm5dYq3XToLkkLv5L2
        pIIVOFMDG+KESnAFV7l2c+cnzRMW0+b6f8mR1CJzZuxVLL6Q02fvLi55/mbSYxECQQDeAw6fiIQX
        GukBI4eMZZt4nscy2o12KyYner3VpoeE+Np2q+Z3pvAMd/aNzQ/W9WaI+NRfcxUJrmfPwIGm63il
        AkEAxCL5HQb2bQr4ByorcMWm/hEP2MZzROV73yF41hPsRC9m66KrheO9HPTJuo3/9s5p+sqGxOlF
        L0NDt4SkosjgGwJAFklyR1uZ/wPJjj611cdBcztlPdqoxssQGnh85BzCj/u3WqBpE2vjvyyvyI5k
        X6zk7S0ljKtt2jny2+00VsBerQJBAJGC1Mg5Oydo5NwD6BiROrPxGo2bpTbu/fhrT8ebHkTz2epl
        U9VQQSQzY1oZMVX8i1m5WUTLPz2yLJIBQVdXqhMCQBGoiuSoSjafUhV7i1cEGpb88h5NBYZzWXGZ
        37sJ5QsW+sJyoNde3xH8vdXhzU7eT82D6X/scw9RZz+/6rCJ4p0=
        -----END RSA PRIVATE KEY-----'''

    @staticmethod
    def gpg_side_effect(*_args, **kwargs):
        """Side effect for gpg key load"""
        kwargs['_out']('fpr:::::::::DD7208BA0A6359F65B906B29CF4AC14A3D109637:')

    @patch('sh.gpg', create=True)
    def test_import_pgp_key_success(self, gpg_mock):
        gpg_mock.side_effect = TestImportPGPKey.gpg_side_effect

        pgp_private_key = TestImportPGPKey.TEST_FAKE_PRIVATE_KEY
        import_pgp_key(
            pgp_private_key=pgp_private_key
        )
        gpg_mock.assert_called_once_with(
            '--import',
            '--fingerprint',
            '--with-colons',
            '--import-options=import-show',
            _in=pgp_private_key,
            _out=Any(IOBase),
            _err_to_out=True,
            _tee='out'
        )

    @patch('sh.gpg', create=True)
    def test_import_pgp_key_gpg_import_fail(self, gpg_mock):
        gpg_mock.side_effect = sh.ErrorReturnCode('gpg', b'mock stdout', b'mock error')

        pgp_private_key = TestImportPGPKey.TEST_FAKE_PRIVATE_KEY

        with self.assertRaisesRegex(
            RuntimeError,
            re.compile(
                r"Error importing pgp private key:"
                r".*RAN: gpg"
                r".*STDOUT:"
                r".*mock stdout"
                r".*STDERR:"
                r".*mock error",
                re.DOTALL
            )
        ):
            import_pgp_key(
                pgp_private_key=pgp_private_key
            )

        gpg_mock.assert_called_once_with(
            '--import',
            '--fingerprint',
            '--with-colons',
            '--import-options=import-show',
            _in=pgp_private_key,
            _out=Any(IOBase),
            _err_to_out=True,
            _tee='out'
        )

    @patch('sh.gpg', create=True)
    def test_import_pgp_key_fail_get_fingerprint(self, gpg_mock):
        pgp_private_key = TestImportPGPKey.TEST_FAKE_PRIVATE_KEY

        with self.assertRaisesRegex(
            RuntimeError,
            re.compile(
                r"Error getting PGP fingerprint for PGP key"
                r" to sign container image\(s\) with. See stdout and stderr for more info.",
                re.DOTALL
            )
        ):
            import_pgp_key(
                pgp_private_key=pgp_private_key
            )

        gpg_mock.assert_called_once_with(
            '--import',
            '--fingerprint',
            '--with-colons',
            '--import-options=import-show',
            _in=pgp_private_key,
            _out=Any(IOBase),
            _err_to_out=True,
            _tee='out'
        )

class TestDetachSignWithPGPKey(BaseTestCase):
    @patch('sh.gpg', create=True)
    def test_detach_sign_with_pgp_key_success(self, gpg_mock):

        detach_sign_with_pgp_key(
            file_to_sign_path='/mock/file-to-detach-sign.txt',
            pgp_private_key_fingerprint='355C137F36628365E42F6D79AF9B11B55E8305E1',
            output_signature_path='/mock/file-to-detach-sign.sig'
        )

        gpg_mock.assert_called_once_with(
            '--armor',
            '--local-user',
            '355C137F36628365E42F6D79AF9B11B55E8305E1',
            '--output',
            '/mock/file-to-detach-sign.sig',
            '--detach-sign',
            '/mock/file-to-detach-sign.txt'
        )

    @patch('sh.gpg', create=True)
    def test_detach_sign_with_pgp_key_failure(self, gpg_mock):
        gpg_mock.side_effect = sh.ErrorReturnCode('gpg', b'mock stdout', b'mock error')

        with self.assertRaisesRegex(
            RuntimeError,
            re.compile(
                r'Error performing detached signature of file \(/mock/file-to-detach-sign.txt\)'
                r' with PGP key \(355C137F36628365E42F6D79AF9B11B55E8305E1\):'
                r".*RAN: gpg"
                r".*STDOUT:"
                r".*mock stdout"
                r".*STDERR:"
                r".*mock error",
                re.DOTALL
            )
        ):
            detach_sign_with_pgp_key(
                file_to_sign_path='/mock/file-to-detach-sign.txt',
                pgp_private_key_fingerprint='355C137F36628365E42F6D79AF9B11B55E8305E1',
                output_signature_path='/mock/file-to-detach-sign.sig'
            )

        gpg_mock.assert_called_once_with(
            '--armor',
            '--local-user',
            '355C137F36628365E42F6D79AF9B11B55E8305E1',
            '--output',
            '/mock/file-to-detach-sign.sig',
            '--detach-sign',
            '/mock/file-to-detach-sign.txt'
        )

class TestExportPGPKey(BaseTestCase):

    TEST_FAKE_FINGERPRINT = 'DEB72DAF711EF3586266C98D9D65E8ECADD724B4'

    TEST_FAKE_PUBLIC_KEY = '''\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mQENBGDKRgEBCACyVB9ZZodlYBf5Fw5iWCohDyjFkypQjRFVaQz8XbvOzrU0reNy
fQpXFzEiNcGxhUHDM538Zv/adIOWMj3EvFTEuCY9I5DSfi4A8AfjSxLMUVSJ85f9
rmHWKoCKoadF7FduKWnGEUk4woVNmhoVqpNMbV6IczNX2vmr7omx4/KgLu4k8WeW
3C0ItJES1fTz9epfynpndXZ9+Iiqiyx8d2cWcVtodv60bJkVIGdLhvyn3/vfSFvf
iSkw4FcKjpXt0k7Kd0e2mv5PHqEYWtkN41p6hDClu3i4EgnFwRjJK9H7lUPcttIJ
hS44e5HhLrihPVK91HZ5Eh0k5jpgQpkNV+ntABEBAAG0O0pvaG4gVGhvbXBraW5z
IChSZWQgSGF0KSAoTm8gY29tbWVudCkgPGp0aG9tcGtpQHJlZGhhdC5jb20+iQFO
BBMBCAA4FiEE3rctr3Ee81hiZsmNnWXo7K3XJLQFAmDKRgECGwMFCwkIBwIGFQoJ
CAsCBBYCAwECHgECF4AACgkQnWXo7K3XJLQANAgAjp5GQALhmhaxLXQJs9tmLzgB
VLOTn6r3z6emlIjjxZYlwLCV+xXZ12Igg+1L5IJo0Nrk2xmcAgp+3WScx29xtkbW
j44jt7eghH4clkSau8RTJWQlyMyzmLVGEDgqBr+YICcnV85NS31ILigvMorKgszP
RVJpG9G+GQRNm7L17sQUSYItX/4ZAbzBr/RhD4IRkgwFxlgrsVujcGN4UMQ2f+uy
b7jY7Gx4o3lKZc+2rFAp6Z9xkChdl69UzJA9Cb7MGNPDE2rFItqXQRUX+4bVOWjy
ffPOuPFqVZJ4989lAfNFh5zitVs2aYGMuWbrRp/AMqBY+2AWw+78/Gv2b6iPfbkB
DQRgykYBAQgAxYk3SS3wk3vGjGF8qAQ6AnMnU2XpqlGt9uTJnChDyzhb964Yxx6+
vN3EqChoAROjyaUQN2GzmUezm1kNDa9sw4KE+a50d5nJqEniz4UvUUTzi9FluTb5
EkRGr1zP0rd1PPWvMB9gil2lie4d7O9iwYT5KUDhiEB+LH8GvI0YENcYDN46CBvG
Qb89qhttXYp22DIll3mQPoabdmWFYD1xYRGoEwP/Zjkx+9MPOASStD4oaULkHoDn
qNMQNrVEY9YQ6JrG55a2XTLEcBaWmLyGnytRjVSs0xp8fzZCwTDgCkcoyAnNrQ7k
wiJJV3fvHP5i6meEdPdRSR4Bwc173ZzUdQARAQABiQE2BBgBCAAgFiEE3rctr3Ee
81hiZsmNnWXo7K3XJLQFAmDKRgECGwwACgkQnWXo7K3XJLS24AgAoGke/g08/FEz
Fml1e81zYEB6+GyggQS365JA16Ev+nvOQbjEk55PvZFfNMbsI0T7IamweX3BzlPf
Sb8HZsjJp0SNlhu7I1P4X3nMa2OofSEE8qd3ptEmAWqewhW+DqGDA23ccAO+VDSj
UvF+nhrYFM/BJY/SE5QyDcDRuJUG3hEAN0FPGSEbWZoAgPpQKdmwLwHeTtdadvAH
NNqw3gdbZjlF+RQbggHc4sXtK/H7wk3SU501BgP10GkfsFUekGiAlkAmYAO6gUnQ
1cfRMdTTzuGVc6j/8VshK1aYRR5A3g5rP4ZbNReqQddZmB20qe4ekkFKbRvHX/Em
B8pBNt1QOA==
=u6VF
-----END PGP PUBLIC KEY BLOCK-----
'''


    @staticmethod
    def gpg_side_effect(*_args, **kwargs):
        """Side effect for gpg key load"""
        kwargs['_out']('gpg: WARNING: nothing exported')

    @patch('sh.gpg', create=True)
    def test_export_pgp_public_key_success(self, gpg_mock):
        gpg_mock.side_effect = TestExportPGPKey.TEST_FAKE_PUBLIC_KEY

        pgp_private_key_fingerprint = TestExportPGPKey.TEST_FAKE_FINGERPRINT
        
        export_pgp_public_key(
            pgp_private_key_fingerprint=pgp_private_key_fingerprint
        )
        gpg_mock.assert_called_once_with(
            '--armor',
            '--export',
            pgp_private_key_fingerprint,
            _out=Any(IOBase),
            _err_to_out=False,
            _tee='out'
        )
        

    @patch('sh.gpg', create=True)
    def test_export_pgp_public_key_failure(self, gpg_mock):
        gpg_mock.side_effect = sh.ErrorReturnCode('gpg', b'mock stdout', b'mock error')
        with self.assertRaisesRegex(
            RuntimeError,
            re.compile(
                r'Error exporting pgp public key:'
                r".*RAN: gpg"
                r".*STDOUT:"
                r".*mock stdout"
                r".*STDERR:"
                r".*mock error",
                re.DOTALL
            )
        ):
            pgp_private_key_fingerprint = TestExportPGPKey.TEST_FAKE_FINGERPRINT
            export_pgp_public_key(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint
            )

        gpg_mock.assert_called_once_with(
            '--armor',
            '--export',
            pgp_private_key_fingerprint,
            _out=Any(IOBase),
            _err_to_out=False,
            _tee='out'
        )


"""A mixin class designed to add Maven test reporting functionality to
MavenGeneric based StepImplementers.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key            | Required? | Default     | Description
-----------------------------|-----------|-------------|-----------
`pom-file`                   | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                 | No        | `True`      | Disables TLS Verification if set to False
`maven-phases-and-goals`     | Yes       |             | List of maven phases and/or goals to execute.
`maven-profiles`             | No        | `[]`        | List of maven profiles to use.
"""# pylint: disable=line-too-long

import os

from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.maven import \
    get_plugin_configuration_absolute_path_values
from ploigos_step_runner.utils.xml import \
    aggregate_xml_element_attribute_values


class MavenTestReportingMixin:
    """A mixin class designed to add Maven test reporting functionality to
    MavenGeneric based StepImplementers.
    """

    SUREFIRE_PLUGIN_NAME = 'maven-surefire-plugin'
    SUREFIRE_PLUGIN_REPORTS_DIR_CONFIG_NAME = 'reportsDirectory'
    FAILSAFE_PLUGIN_NAME = 'maven-failsafe-plugin'
    FAILSAFE_PLUGIN_REPORTS_DIR_CONFIG_NAME = 'reportsDirectory'
    SUREFIRE_PLUGIN_DEFAULT_REPORTS_DIR = 'target/surefire-reports'
    FAILSAFE_PLUGIN_DEFAULT_REPORTS_DIR = 'target/failsafe-reports'
    TESTSUITE_EVIDENCE_ATTRIBUTES = ["time", "tests", "errors", "skipped", "failures"]

    def _attempt_get_test_report_directory(
        self,
        plugin_name,
        configuration_key,
        default,
        require_phase_execution_config=False
    ):
        """Does it's darndest to dynamically determine the test report directory.

        Parameters
        ----------
        plugin_name : str
            Name of the Maven plugin to look for test report directory configuration.
        configuration_key : str
            Maven plugin configuration to look for the test directory path.
        default : str
            Value to use if can't find any user configured configuration.
        require_phase_execution_config : str
            True if the user supplied configuration via the pom should be for
            the specified phase or goals.
            False if the user supplied configuration via the pom does not
            need to be specific for the phase and goal.

        Returns
        -------
        str
            Determined test reports directory path.

        Raises
        ------
        StepRunnerException
            If can not find the given plugin to get configuration from.
        """
        test_report_dir = None
        print(
            'Attempt to get test report directory configuration'
            f' ({configuration_key}) for'
            f' maven test plugin ({plugin_name}).'
        )
        try:
            test_report_dirs = get_plugin_configuration_absolute_path_values(
                plugin_name=plugin_name,
                configuration_key=configuration_key,
                work_dir_path=self.work_dir_path,
                pom_file=self.get_value('pom-file'),
                profiles=self.get_value('maven-profiles'),
                phases_and_goals=self.maven_phases_and_goals,
                require_phase_execution_config=require_phase_execution_config
            )

            # if found at least one test report dir
            # else plugin exists but could not find config, use default
            if test_report_dirs:
                if len(test_report_dirs) > 1:
                    print(
                        'WARNING: In best attempt to dynamically determine where the the test'
                        ' report directory is, we were to successful and found more then one.'
                        ' This is not wholly unexpected because there is enumerable maven plugins,'
                        ' and enumerable ways to configure them.'
                        ' Randomly picking first match and hoping it is correct.'
                        ' Rather then relying on this step implementer to try and figure out'
                        ' where the test reports are you can configure it manually via the'
                        ' step implementer config (test-reports-dir).'
                    )

                test_report_dir = test_report_dirs[0]
            else:
                print(
                    'Did not find test report directory configuration'
                    f' ({configuration_key}) for maven test plugin ({plugin_name}),'
                    f' using default ({default}).'
                )
                test_report_dir = default

        except RuntimeError as error:
            # NOTE: this should only happen if couldn't find the plugin
            raise StepRunnerException(
                f'Error getting configuration ({configuration_key}) from'
                f' maven plugin ({plugin_name}): {error}'
            ) from error

        return test_report_dir

    @staticmethod
    def _gather_evidence_from_test_report_directory_testsuite_elements(
        step_result,
        test_report_dir
    ):
        """Given a test report directory containing XML files with 'testsuite' xml elements
        collects evidence from those files and elements.

        Parameters
        ----------
        step_result : StepResult
            StepResult to add the evidence to.
        test_report_dir : str
            Directory to search for 'testsuite' xml elements in to collect evidence from.
        """
        if os.path.exists(test_report_dir):
            test_report_evidence_attributes = MavenTestReportingMixin.TESTSUITE_EVIDENCE_ATTRIBUTES
            test_report_evidence_element = 'testsuite'

            not_found_attribs = []
            report_results = aggregate_xml_element_attribute_values(
                xml_file_paths=test_report_dir,
                xml_element=test_report_evidence_element,
                attribs=test_report_evidence_attributes
            )

            for attribute in test_report_evidence_attributes:
                if attribute in report_results:
                    step_result.add_evidence(
                        name=attribute,
                        value=report_results[attribute]
                    )
                else:
                    not_found_attribs.append(attribute)
            if not_found_attribs:
                step_result.message += "\nWARNING: could not find expected evidence" \
                    f" attributes ({not_found_attribs}) on xml element" \
                    f" ({test_report_evidence_element}) in test report" \
                    f" directory ({test_report_dir})."
        else:
            step_result.message += f"\nWARNING: test report directory ({test_report_dir})" \
                " does not exist to gather evidence from"



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os
import re
from io import IOBase
from pathlib import Path
from unittest.mock import patch

import mock
import sh
from ploigos_step_runner import StepResult
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.step_implementers.sign_container_image import \
    PodmanSign
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any


class TestStepImplementerSignContainerImagePodman(BaseStepImplementerTestCase):
    TEST_FAKE_PRIVATE_KEY = '''
        -----BEGIN RSA PRIVATE KEY-----
        MIICXAIBAAKBgQCqGKukO1De7zhZj6+H0qtjTkVxwTCpvKe4eCZ0FPqri0cb2JZfXJ/DgYSF6vUp
        wmJG8wVQZKjeGcjDOL5UlsuusFncCzWBQ7RKNUSesmQRMSGkVb1/3j+skZ6UtW+5u09lHNsj6tQ5
        1s1SPrCBkedbNf0Tp0GbMJDyR4e9T04ZZwIDAQABAoGAFijko56+qGyN8M0RVyaRAXz++xTqHBLh
        3tx4VgMtrQ+WEgCjhoTwo23KMBAuJGSYnRmoBZM3lMfTKevIkAidPExvYCdm5dYq3XToLkkLv5L2
        pIIVOFMDG+KESnAFV7l2c+cnzRMW0+b6f8mR1CJzZuxVLL6Q02fvLi55/mbSYxECQQDeAw6fiIQX
        GukBI4eMZZt4nscy2o12KyYner3VpoeE+Np2q+Z3pvAMd/aNzQ/W9WaI+NRfcxUJrmfPwIGm63il
        AkEAxCL5HQb2bQr4ByorcMWm/hEP2MZzROV73yF41hPsRC9m66KrheO9HPTJuo3/9s5p+sqGxOlF
        L0NDt4SkosjgGwJAFklyR1uZ/wPJjj611cdBcztlPdqoxssQGnh85BzCj/u3WqBpE2vjvyyvyI5k
        X6zk7S0ljKtt2jny2+00VsBerQJBAJGC1Mg5Oydo5NwD6BiROrPxGo2bpTbu/fhrT8ebHkTz2epl
        U9VQQSQzY1oZMVX8i1m5WUTLPz2yLJIBQVdXqhMCQBGoiuSoSjafUhV7i1cEGpb88h5NBYZzWXGZ
        37sJ5QsW+sJyoNde3xH8vdXhzU7eT82D6X/scw9RZz+/6rCJ4p0=
        -----END RSA PRIVATE KEY-----'''

    @staticmethod
    def generate_config():
        return {
            'signer-pgp-private-key': \
                TestStepImplementerSignContainerImagePodman.TEST_FAKE_PRIVATE_KEY
        }

    @staticmethod
    def create_podman_image_sign_side_effect(num_signatures=1):
        def podman_image_sign_side_effect(*args, **kwargs):
            if args[0] == 'sign':
                # determine the directory for the signature
                for arg in args:
                    match = re.match(r'^--directory=(?P<image_signature_directory>.+)$', arg)
                    if match:
                        image_signatures_directory =match.groupdict()['image_signature_directory']
                        break

                # touch mock signature file
                mock_image_signature_dir_path = os.path.join(
                    image_signatures_directory,
                    'mock/sig/path'
                )
                os.makedirs(mock_image_signature_dir_path)

                for i in range(num_signatures):
                    mock_image_signature_path = os.path.join(
                        mock_image_signature_dir_path,
                        f'signature-{i}'
                    )
                    Path(mock_image_signature_path).touch()

        return podman_image_sign_side_effect

    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=PodmanSign,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_step_implementer_config_defaults(self):
        defaults = PodmanSign.step_implementer_config_defaults()
        expected_defaults = {
            'src-tls-verify': 'true'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = PodmanSign._required_config_or_result_keys()
        expected_required_keys = [
            ['signer-pgp-private-key', 'container-image-signer-pgp-private-key'],
            ['container-image-push-tag', 'container-image-tag']
        ]
        self.assertEqual(required_keys, expected_required_keys)

    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.container_registries_login')
    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.import_pgp_key')
    @patch.object(PodmanSign, '_PodmanSign__sign_image')
    def test_run_step_pass(
        self,
        sign_image_mock,
        import_pgp_key_mock,
        container_registries_login
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            pgp_private_key_fingerprint = 'abc123'
            step_config = TestStepImplementerSignContainerImagePodman.generate_config()
            container_image_tag = 'does/not/matter:v0.42.0'
            signature_name = 'does/not/matter/signature-0'

            # Previous (fake) results
            artifact_config = {
                'container-image-tag': {'value': container_image_tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            def import_pgp_key_side_effect(pgp_private_key):
                return pgp_private_key_fingerprint
            import_pgp_key_mock.side_effect = import_pgp_key_side_effect

            def sign_image_side_effect(
                pgp_private_key_fingerprint,
                image_signatures_directory,
                container_image_tag
            ):
                return os.path.join(image_signatures_directory, signature_name)
            sign_image_mock.side_effect = sign_image_side_effect

            # Actual results
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='sign-container-image',
                implementer='PodmanSign',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()
            import_pgp_key_mock.assert_called_once_with(
                pgp_private_key=step_config['signer-pgp-private-key']
            )
            sign_image_mock.assert_called_once_with(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                image_signatures_directory= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature'
                ),
                container_image_tag=container_image_tag
            )

            container_registries_login.assert_called_once_with(
                registries=None,
                containers_config_tls_verify=True,
                container_command_short_name='podman'
            )
            expected_step_result = StepResult(
                step_name='sign-container-image',
                sub_step_name='PodmanSign',
                sub_step_implementer_name='PodmanSign'
            )
            expected_step_result.add_artifact(
                name='container-image-signed-tag',
                value=container_image_tag,
            )
            expected_step_result.add_artifact(
                name='container-image-signature-file-path',
                value= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature',
                    signature_name
                )
            )
            expected_step_result.add_artifact(
                name='container-image-signature-name',
                value=signature_name
            )
            expected_step_result.add_artifact(
                name='container-image-signature-signer-private-key-fingerprint',
                value=pgp_private_key_fingerprint
            )
            self.assertEqual(expected_step_result, result)

    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.upload_file')
    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.import_pgp_key')
    @patch.object(PodmanSign, '_PodmanSign__sign_image')
    def test_run_step_pass_with_signature_upload_to_file(
        self,
        sign_image_mock,
        import_pgp_key_mock,
        upload_file_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            pgp_private_key_fingerprint = 'abc123'
            step_config = TestStepImplementerSignContainerImagePodman.generate_config()
            step_config['container-image-signature-destination-url'] = '/mock/container-sigs'
            container_image_tag = 'does/not/matter:v0.42.0'
            signature_name = 'does/not/matter/signature-0'

            # Previous (fake) results
            artifact_config = {
                'container-image-tag': {'value': container_image_tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            def import_pgp_key_side_effect(pgp_private_key):
                return pgp_private_key_fingerprint
            import_pgp_key_mock.side_effect = import_pgp_key_side_effect

            def sign_image_side_effect(
                    pgp_private_key_fingerprint,
                    image_signatures_directory,
                    container_image_tag
            ):
                return os.path.join(image_signatures_directory, signature_name)
            sign_image_mock.side_effect = sign_image_side_effect

            upload_file_mock.return_value = "mock upload results"

            # Actual results
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='sign-container-image',
                implementer='PodmanSign',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()
            import_pgp_key_mock.assert_called_once_with(
                pgp_private_key=step_config['signer-pgp-private-key']
            )
            sign_image_mock.assert_called_once_with(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                image_signatures_directory= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature'
                ),
                container_image_tag=container_image_tag
            )
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri='/mock/container-sigs/does/not/matter/signature-0',
                username=None,
                password=None
            )

            expected_step_result = StepResult(
                step_name='sign-container-image',
                sub_step_name='PodmanSign',
                sub_step_implementer_name='PodmanSign'
            )
            expected_step_result.add_artifact(
                name='container-image-signed-tag',
                value=container_image_tag,
            )
            expected_step_result.add_artifact(
                name='container-image-signature-file-path',
                value= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature',
                    signature_name
                )
            )
            expected_step_result.add_artifact(
                name='container-image-signature-name',
                value=signature_name
            )
            expected_step_result.add_artifact(
                name='container-image-signature-signer-private-key-fingerprint',
                value=pgp_private_key_fingerprint
            )
            expected_step_result.add_artifact(
                name='container-image-signature-uri',
                description='URI of the uploaded container image signature',
                value='/mock/container-sigs/does/not/matter/signature-0'
            )
            expected_step_result.add_artifact(
                name='container-image-signature-upload-results',
                description='Results of uploading the container image signature' \
                            ' to the given destination.',
                value='mock upload results'
            )

            self.assertEqual(expected_step_result, result)

    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.upload_file')
    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.import_pgp_key')
    @patch.object(PodmanSign, '_PodmanSign__sign_image')
    def test_run_step_pass_with_signature_upload_to_remote_with_auth(
        self,
        sign_image_mock,
        import_pgp_key_mock,
        upload_file_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            pgp_private_key_fingerprint = 'abc123'
            step_config = TestStepImplementerSignContainerImagePodman.generate_config()
            step_config['container-image-signature-destination-url'] = \
                'https://ploigos.com/mock/container-sigs'
            step_config['container-image-signature-destination-username'] = 'test-user'
            step_config['container-image-signature-destination-password'] = 'test-pass'
            container_image_tag = 'does/not/matter:v0.42.0'
            signature_name = 'does/not/matter/signature-0'

            # Previous (fake) results
            artifact_config = {
                'container-image-tag': {'value': container_image_tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            def import_pgp_key_side_effect(pgp_private_key):
                return pgp_private_key_fingerprint
            import_pgp_key_mock.side_effect = import_pgp_key_side_effect

            def sign_image_side_effect(
                    pgp_private_key_fingerprint,
                    image_signatures_directory,
                    container_image_tag
            ):
                return os.path.join(image_signatures_directory, signature_name)
            sign_image_mock.side_effect = sign_image_side_effect

            upload_file_mock.return_value = "mock upload results"

            # Actual results
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='sign-container-image',
                implementer='PodmanSign',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()
            import_pgp_key_mock.assert_called_once_with(
                pgp_private_key=step_config['signer-pgp-private-key']
            )
            sign_image_mock.assert_called_once_with(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                image_signatures_directory= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature'
                ),
                container_image_tag=container_image_tag
            )
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri='https://ploigos.com/mock/container-sigs/does/not/matter/signature-0',
                username='test-user',
                password='test-pass'
            )

            expected_step_result = StepResult(
                step_name='sign-container-image',
                sub_step_name='PodmanSign',
                sub_step_implementer_name='PodmanSign'
            )
            expected_step_result.add_artifact(
                name='container-image-signed-tag',
                value=container_image_tag,
            )
            expected_step_result.add_artifact(
                name='container-image-signed-tag',
                value=container_image_tag,
            )
            expected_step_result.add_artifact(
                name='container-image-signature-file-path',
                value= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature',
                    signature_name
                )
            )
            expected_step_result.add_artifact(
                name='container-image-signature-name',
                value=signature_name
            )
            expected_step_result.add_artifact(
                name='container-image-signature-signer-private-key-fingerprint',
                value=pgp_private_key_fingerprint
            )
            expected_step_result.add_artifact(
                name='container-image-signature-uri',
                description='URI of the uploaded container image signature',
                value='https://ploigos.com/mock/container-sigs/does/not/matter/signature-0'
            )
            expected_step_result.add_artifact(
                name='container-image-signature-upload-results',
                description='Results of uploading the container image signature' \
                            ' to the given destination.',
                value='mock upload results'
            )

            self.assertEqual(expected_step_result, result)

    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.upload_file')
    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.import_pgp_key')
    @patch.object(PodmanSign, '_PodmanSign__sign_image')
    def test_run_step_pass_with_signature_upload_to_remote_with_auth_failure(
        self,
        sign_image_mock,
        import_pgp_key_mock,
        upload_file_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            pgp_private_key_fingerprint = 'abc123'
            step_config = TestStepImplementerSignContainerImagePodman.generate_config()
            step_config['container-image-signature-destination-url'] = \
                'https://ploigos.com/mock/container-sigs'
            step_config['container-image-signature-destination-username'] = 'test-user'
            step_config['container-image-signature-destination-password'] = 'test-pass'
            container_image_tag = 'does/not/matter:v0.42.0'
            signature_name = 'does/not/matter/signature-0'

            # Previous (fake) results
            artifact_config = {
                'container-image-tag': {'value': container_image_tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            def import_pgp_key_side_effect(pgp_private_key):
                return pgp_private_key_fingerprint
            import_pgp_key_mock.side_effect = import_pgp_key_side_effect

            def sign_image_side_effect(
                    pgp_private_key_fingerprint,
                    image_signatures_directory,
                    container_image_tag
            ):
                return os.path.join(image_signatures_directory, signature_name)
            sign_image_mock.side_effect = sign_image_side_effect

            upload_file_mock.side_effect = RuntimeError('mock upload error')

            # Actual results
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='sign-container-image',
                implementer='PodmanSign',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            import_pgp_key_mock.assert_called_once_with(
                pgp_private_key=step_config['signer-pgp-private-key']
            )
            sign_image_mock.assert_called_once_with(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                image_signatures_directory= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature'
                ),
                container_image_tag=container_image_tag
            )
            upload_file_mock.assert_called_once_with(
                file_path=mock.ANY,
                destination_uri='https://ploigos.com/mock/container-sigs/does/not/matter/signature-0',
                username='test-user',
                password='test-pass'
            )

            expected_step_result = StepResult(
                step_name='sign-container-image',
                sub_step_name='PodmanSign',
                sub_step_implementer_name='PodmanSign'
            )
            expected_step_result.add_artifact(
                name='container-image-signed-tag',
                value=container_image_tag,
            )
            expected_step_result.add_artifact(
                name='container-image-signature-file-path',
                value= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature',
                    signature_name
                )
            )
            expected_step_result.add_artifact(
                name='container-image-signature-name',
                value=signature_name
            )
            expected_step_result.add_artifact(
                name='container-image-signature-signer-private-key-fingerprint',
                value=pgp_private_key_fingerprint
            )
            expected_step_result.add_artifact(
                name='container-image-signature-uri',
                description='URI of the uploaded container image signature',
                value='https://ploigos.com/mock/container-sigs/does/not/matter/signature-0'
            )
            expected_step_result.success = False
            expected_step_result.message = 'mock upload error'

            self.assertEqual(expected_step_result, result)

    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.import_pgp_key')
    @patch.object(PodmanSign, '_PodmanSign__sign_image')
    def test_run_step_fail_import_pgp_key(self, sign_image_mock, import_pgp_key_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = TestStepImplementerSignContainerImagePodman.generate_config()
            container_image_tag = 'does/not/matter:v0.42.0'
            signature_name = 'does/not/matter/signature-0'

            # Previous (fake) results
            artifact_config = {
                'container-image-tag': {'value': container_image_tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            import_pgp_key_mock.side_effect = RuntimeError('mock error importing pgp key')

            def sign_image_side_effect(
                    pgp_private_key_fingerprint,
                    image_signatures_directory,
                    container_image_tag
            ):
                return os.path.join(image_signatures_directory, signature_name)

            sign_image_mock.side_effect = sign_image_side_effect

            # Actual results
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='sign-container-image',
                implementer='PodmanSign',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()
            import_pgp_key_mock.assert_called_once_with(
                pgp_private_key=step_config['signer-pgp-private-key']
            )

            expected_step_result = StepResult(
                step_name='sign-container-image',
                sub_step_name='PodmanSign',
                sub_step_implementer_name='PodmanSign'
            )
            expected_step_result.success = False
            expected_step_result.message = 'mock error importing pgp key'

            self.assertEqual(expected_step_result, result)

    @patch('ploigos_step_runner.step_implementers.sign_container_image.podman_sign.import_pgp_key')
    @patch.object(PodmanSign, '_PodmanSign__sign_image')
    def test_run_step_fail_sign_image(self, sign_image_mock, import_pgp_key_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            pgp_private_key_fingerprint = 'abc123'
            step_config = TestStepImplementerSignContainerImagePodman.generate_config()
            container_image_tag = 'does/not/matter:v0.42.0'

            # Previous (fake) results
            artifact_config = {
                'container-image-tag': {'value': container_image_tag}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            def import_pgp_key_side_effect(pgp_private_key):
                return pgp_private_key_fingerprint
            import_pgp_key_mock.side_effect = import_pgp_key_side_effect

            sign_image_mock.side_effect = StepRunnerException('mock error signing image')

            # Actual results
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='sign-container-image',
                implementer='PodmanSign',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()
            import_pgp_key_mock.assert_called_once_with(
                pgp_private_key=step_config['signer-pgp-private-key']
            )
            sign_image_mock.assert_called_once_with(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                image_signatures_directory= os.path.join(
                    parent_work_dir_path,
                    'sign-container-image/image-signature'
                ),
                container_image_tag=container_image_tag
            )

            expected_step_result = StepResult(
                step_name='sign-container-image',
                sub_step_name='PodmanSign',
                sub_step_implementer_name='PodmanSign'
            )
            expected_step_result.add_artifact(
                name='container-image-signature-signer-private-key-fingerprint',
                value=pgp_private_key_fingerprint
            )
            expected_step_result.success = False
            expected_step_result.message = 'mock error signing image'

            self.assertEqual(expected_step_result, result)

    @patch('sh.podman', create=True)
    def test___sign_image_success(self, podman_mock):
        with TempDirectory() as temp_dir:
            pgp_private_key_fingerprint = 'abc123'
            image_signatures_directory = os.path.join(temp_dir.path, 'signatures')
            container_image_tag = 'does/not/matter:v0.42.0'

            podman_mock.image.side_effect = TestStepImplementerSignContainerImagePodman.\
                create_podman_image_sign_side_effect()

            PodmanSign._PodmanSign__sign_image(
                pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                image_signatures_directory=image_signatures_directory,
                container_image_tag=container_image_tag
            )

            podman_mock.image.assert_called_once_with(
                'sign',
                f'--sign-by={pgp_private_key_fingerprint}',
                f'--directory={image_signatures_directory}',
                f'docker://{container_image_tag}',
                _out=Any(IOBase),
                _err_to_out=True,
                _tee='out'
            )

    @patch('sh.podman', create=True)
    def test___sign_image_podman_image_sign_fail(self, podman_mock):
        with TempDirectory() as temp_dir:
            pgp_private_key_fingerprint = 'abc123'
            image_signatures_directory = os.path.join(temp_dir.path, 'signatures')
            container_image_tag = 'does/not/matter:v0.42.0'

            podman_mock.image.side_effect = sh.ErrorReturnCode('podman', b'mock stdout', b'mock error signing image')

            with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    rf"Error signing image \({container_image_tag}\):"
                    r".*RAN: podman"
                    r".*STDOUT:"
                    r".*mock stdout"
                    r".*STDERR:"
                    r".*mock error signing image",
                    re.DOTALL
                )
            ):
                PodmanSign._PodmanSign__sign_image(
                    pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                    image_signatures_directory=image_signatures_directory,
                    container_image_tag=container_image_tag
                )

            podman_mock.image.assert_called_once_with(
                'sign',
                f'--sign-by={pgp_private_key_fingerprint}',
                f'--directory={image_signatures_directory}',
                f'docker://{container_image_tag}',
                _out=Any(IOBase),
                _err_to_out=True,
                _tee='out'
            )

    @patch('sh.podman', create=True)
    def test___sign_image_podman_image_sign_to_many_sigs(self, podman_mock):
        with TempDirectory() as temp_dir:
            pgp_private_key_fingerprint = 'abc123'
            image_signatures_directory = os.path.join(temp_dir.path, 'signatures')
            container_image_tag = 'does/not/matter:v0.42.0'

            podman_mock.image.side_effect = TestStepImplementerSignContainerImagePodman.\
                create_podman_image_sign_side_effect(num_signatures=2)

            with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Unexpected number of signature files, expected 1: \['.*', '.*'\]",
                    re.DOTALL
                )
            ):
                PodmanSign._PodmanSign__sign_image(
                    pgp_private_key_fingerprint=pgp_private_key_fingerprint,
                    image_signatures_directory=image_signatures_directory,
                    container_image_tag=container_image_tag
                )

            podman_mock.image.assert_called_once_with(
                'sign',
                f'--sign-by={pgp_private_key_fingerprint}',
                f'--directory={image_signatures_directory}',
                f'docker://{container_image_tag}',
                _out=Any(IOBase),
                _err_to_out=True,
                _tee='out'
            )



"""
Shared utils for npm operations.
"""
import sys
import sh
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.io import \
    create_sh_redirect_to_multiple_streams_fn_callback

def run_npm(npm_output_file_path, npm_args):
    """
    Run an npm command

    Parameters
    ----------
    npm_output_file_path:
        String
    npm_args:
        Commandline arguments to npm
    """
    try:
        with open(npm_output_file_path, 'w', encoding='utf-8') as npm_output_file:
            out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                sys.stdout,
                npm_output_file
            ])
            err_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                sys.stderr,
                npm_output_file
            ])

            sh.npm( # pylint: disable=no-member
                npm_args,
                _out=out_callback,
                _err=err_callback
            )
    except sh.ErrorReturnCode as error:
        raise StepRunnerException(
            f"Error running npm. {error}"
        ) from error



"""`StepImplementers` for the `unit-test` step.
"""

from ploigos_step_runner.step_implementers.unit_test.maven import Maven
from ploigos_step_runner.step_implementers.unit_test.maven_test import \
    MavenTest
from ploigos_step_runner.step_implementers.unit_test.npm_test import NpmTest



"""
Shared utilities for dealing with Python reflection.
"""

def import_and_get_class(module_name, class_name):
    """Dynamically loads a class from a given module.

    Parameters
    ----------
    module_name : str
        Name of the module to load the class from.
    class_name : str
        Name of the class to load from the given module.

    Returns
    -------
    class or None
        The class from the the given module, or none if given class does not exist in given module.
    """

    try:
        module = __import__(module_name, fromlist=[class_name])
        clazz = getattr(module, class_name)
    except (AttributeError, ModuleNotFoundError):
        clazz = None

    return clazz



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os
from unittest.mock import patch

import sh
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from ploigos_step_runner.step_implementers.validate_environment_configuration import \
    Configlint
from ploigos_step_runner import StepResult


class TestStepImplementerConfiglint(BaseStepImplementerTestCase):
    @staticmethod
    def create_config_lint_side_effect(
            config_lint_stdout="",
            config_lint_stderr="",
            config_lint_fail=False
    ):
        def config_lint_side_effect(*args, **kwargs):
            kwargs['_out'](config_lint_stdout)
            kwargs['_err'](config_lint_stderr)

            if config_lint_fail:
                raise sh.ErrorReturnCode_255(
                    'config-lint',
                    bytes(config_lint_stdout, 'utf-8'),
                    bytes(config_lint_stderr, 'utf-8')
                )

        return config_lint_side_effect

    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Configlint,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    # TESTS FOR configuration checks
    def test_step_implementer_config_defaults(self):
        defaults = Configlint.step_implementer_config_defaults()
        expected_defaults = {
            'rules': './config-lint.rules'
        }
        self.assertEqual(expected_defaults, defaults)

    def test__required_config_or_result_keys(self):
        required_keys = Configlint._required_config_or_result_keys()
        expected_required_keys = [
            'configlint-yml-path'
        ]
        self.assertEqual(required_keys, expected_required_keys)

    @patch('sh.config_lint', create=True)
    def test_run_step_fail_bad_yml_path(self, config_lint_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            test_file_name = 'rules'
            test_file_path = os.path.join(temp_dir.path, test_file_name)
            temp_dir.write(test_file_path, b'ignored')
            step_config = {
                'rules': test_file_path,
                'configlint-yml-path': 'invalid_file'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='Configlint',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='Configlint',
                sub_step_implementer_name='Configlint'
            )

            expected_step_result.success = False
            expected_step_result.message = 'File specified in configlint-yml-path not found: invalid_file'
            self.assertEqual(expected_step_result, result)

    @patch('sh.config_lint', create=True)
    def test_run_step_fail_bad_rule_path(self, config_lint_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            test_file_name = 'rules'
            test_file_path = os.path.join(temp_dir.path, test_file_name)
            temp_dir.write(test_file_path, b'ignored')
            step_config = {
                'configlint-yml-path': test_file_path,
                'rules': 'invalid_file'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='Configlint',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='Configlint',
                sub_step_implementer_name='Configlint'
            )

            expected_step_result.success = False
            expected_step_result.message = 'File specified in rules not found: invalid_file'
            self.assertEqual(expected_step_result, result)

    @patch('sh.config_lint', create=True)
    def test_run_step_pass_prev_step(self, config_lint_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            test_file_name = 'rules'
            test_file_path = os.path.join(temp_dir.path, test_file_name)
            temp_dir.write(test_file_path, b'ignored')
            step_config = {
                'rules': test_file_path,
            }

            artifact_config = {
                'configlint-yml-path': {'value': f'{test_file_path}'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='Configlint',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='Configlint',
                sub_step_implementer_name='Configlint'
            )

            expected_step_result.add_artifact(
                name='configlint-result-set',
                value=f'{step_implementer.work_dir_path}/configlint_results_file.txt'
            )
            expected_step_result.add_artifact(
                name='configlint-yml-path',
                value=test_file_path
            )
            self.assertEqual(expected_step_result, result)

    @patch('sh.config_lint', create=True)
    def test_run_step_fail_scan(self, configlint_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            file_to_validate_contents = 'notused'
            temp_dir.write('config-file-to-validate.yml', file_to_validate_contents.encode())
            file_to_validate_file_path = str(os.path.join(temp_dir.path, 'config-file-to-validate.yml'))

            # write config-lint rules file
            configlint_rules_content = 'not used'
            config_lint_rules_file_name = 'config-lint-test-rules.yml'
            temp_dir.write(config_lint_rules_file_name, configlint_rules_content.encode())
            config_lint_rules_file_path = os.path.join(temp_dir.path, config_lint_rules_file_name)

            step_config = {
                'configlint-yml-path': file_to_validate_file_path,
                'rules': config_lint_rules_file_path
            }

            configlint_mock.side_effect = \
                TestStepImplementerConfiglint.create_config_lint_side_effect(
                    config_lint_fail=True
                )

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='Configlint',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='Configlint',
                sub_step_implementer_name='Configlint'
            )

            expected_step_result.success = False
            expected_step_result.message = 'Failed config-lint scan.'
            expected_step_result.add_artifact(
                name='configlint-result-set',
                value=f'{step_implementer.work_dir_path}/configlint_results_file.txt'
            )
            expected_step_result.add_artifact(
                name='configlint-yml-path',
                value=file_to_validate_file_path
            )
            self.assertEqual(expected_step_result, result)

    @patch('sh.config_lint', create=True)
    def test_run_step_pass(self, config_lint_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            test_file_name = 'file.txt'
            test_file_path = os.path.join(temp_dir.path, test_file_name)
            temp_dir.write(test_file_path, b'ignored')
            step_config = {
                'rules': test_file_path,
                'configlint-yml-path': test_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='Configlint',
                parent_work_dir_path=parent_work_dir_path,
            )

            config_lint_mock.side_effect = sh.ErrorReturnCode('config_lint', b'mock out', b'mock error')
            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='Configlint',
                sub_step_implementer_name='Configlint'
            )

            expected_step_result.success = False
            expected_step_result.message = 'Unexpected Error invoking config-lint.'
            expected_step_result.add_artifact(
                name='configlint-result-set',
                value=f'{step_implementer.work_dir_path}/configlint_results_file.txt'
            )
            expected_step_result.add_artifact(
                name='configlint-yml-path',
                value=test_file_path
            )
            self.assertEqual(expected_step_result, result)



"""`StepImplementer` for the `package` step using Maven.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key            | Required? | Default | Description
-----------------------------|-----------|---------|-----------
`pom-file`                   | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                 | No        | `True`  | Disables TLS Verification if set to False
`maven-profiles`             | No        | `[]`    | List of maven profiles to use.
`maven-no-transfer-progress` | No        | `True`  | \
                            `True` to suppress the transfer progress of packages maven downloads.
                            `False` to have the transfer progress printed.\
                            See https://maven.apache.org/ref/current/maven-embedder/cli.html
`maven-additional-arguments` | No        | `['-Dmaven.install.skip=true', \
                                             '-Dmaven.test.skip=true' \
                                             '-DskipTests', \
                                             '-DskipITs']` \
                                                   | List of additional arguments to use. \
                                                     Skipping tests by default because assuming \
                                                     a previous step already ran them. \
                                                     Skipping install backs assuming this is \
                                                     running in an ephemeral environment where \
                                                     that would be a waist of time, and also \
                                                     that a previous step ran `package` \
                                                     and `push-artifacts` steps.
`maven-servers`              | No        |         | Dictionary of dictionaries of \
                                                     id, username, password
`maven-repositories`         | No        |         | Dictionary of dictionaries of \
                                                     id, url, snapshots, releases
`maven-mirrors`              | No        |         | Dictionary of dictionaries of \
                                                     id, url, mirror_of
`version`                      | Yes     |         | version to push
`maven-push-artifact-repo-url` | yes     |         | id for the maven servers and mirrors
`maven-push-artifact-repo-id`  | Yes     |         | url for the maven servers and mirrors

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`push-artifacts`    | An array of dictionaries with information on the pushed artifacts
"""

from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementers.shared.maven_generic import \
    MavenGeneric
from ploigos_step_runner.utils.maven import run_maven

DEFAULT_CONFIG = {
    'maven-additional-arguments': [
        '-Dmaven.install.skip=true',
        '-Dmaven.test.skip=true'
        '-DskipTests',
        '-DskipITs'
    ]
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'pom-file',
    'maven-push-artifact-repo-url',
    'maven-push-artifact-repo-id',
    'version'
]

class MavenDeploy(MavenGeneric):
    """`StepImplementer` for the `push-artifacts` step using Maven.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment,
            maven_phases_and_goals=['deploy']
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return {**MavenGeneric.step_implementer_config_defaults(), **DEFAULT_CONFIG}

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self): # pylint: disable=too-many-locals
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # Get config items
        maven_push_artifact_repo_id = self.get_value('maven-push-artifact-repo-id')
        maven_push_artifact_repo_url = self.get_value('maven-push-artifact-repo-url')
        version = self.get_value('version')

        # push the artifacts
        mvn_update_version_output_file_path = self.write_working_file('mvn_versions_set_output.txt')
        mvn_push_artifacts_output_file_path = self.write_working_file('mvn_deploy_output.txt')
        try:
            # update the version before pushing
            # NOTE 1: we know this is weird. But the version in the pom isn't necessarily
            #         the version that was calculated as part of the release and so we need
            #         to update that before doing the maven deploy so the maven deploy will
            #         use the new version.
            #
            # NOTE 2: we tried doing this in the same command as the deploy,
            #         but the pom was already loaded so even though the xml was updated
            #         the deploy still used the old version, hence having to run this
            #         first and independently.
            print("Update maven package version")
            run_maven(
                mvn_output_file_path=mvn_update_version_output_file_path,
                settings_file=self.maven_settings_file,
                pom_file=self.get_value('pom-file'),
                phases_and_goals=['versions:set'],
                additional_arguments=[
                    f'-DnewVersion={version}'
                ]
            )

            # execute maven step (params come from config)
            print("Push packaged maven artifacts")
            self._run_maven_step(
                mvn_output_file_path=mvn_push_artifacts_output_file_path,
                step_implementer_additional_arguments=[
                    '-DaltDeploymentRepository=' \
                    f'{maven_push_artifact_repo_id}::default::{maven_push_artifact_repo_url}'
                ]
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running 'maven deploy' to push artifacts. " \
                f"More details maybe found in 'maven-output' report artifact: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from running maven to update version.",
                name='maven-update-version-output',
                value=mvn_update_version_output_file_path
            )
            step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "push artifacts to repository.",
                name='maven-push-artifacts-output',
                value=mvn_push_artifacts_output_file_path
            )

        return step_result




import os
from unittest.mock import MagicMock, patch

from ploigos_step_runner import StepRunnerException
from ploigos_step_runner.results import StepResult
from ploigos_step_runner.step_implementers.shared import \
    MavenTestReportingMixin
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.step_implementers.shared.maven_test_reporting_mixin.get_plugin_configuration_absolute_path_values")
class TestMavenTestReportingMixin__attempt_get_test_report_directory(BaseStepImplementerTestCase):
    @staticmethod
    def __get_maven_test_reporting_mixin():
        # create mixin object
        maven_test_reporting_mixin = MavenTestReportingMixin()

        # mock work_dir_path
        maven_test_reporting_mixin.work_dir_path = '/mock/work-dir-path'

        # mock get_value

        def get_value_side_effect(key):
            if key == 'pom-file':
                return 'mock-pom.xml'
            elif key == 'maven-profiles':
                return []
            else:
                return None
        maven_test_reporting_mixin.get_value = MagicMock(
            name='get_value',
            side_effect = get_value_side_effect
        )

        # mock maven_phases_and_goals
        maven_test_reporting_mixin.maven_phases_and_goals = []

        return maven_test_reporting_mixin

    def test_one_found_result(self, get_plugin_configuration_absolute_path_values_mock):
        # create object to test against
        maven_test_reporting_mixin = self.__get_maven_test_reporting_mixin()

        # setup mocks
        get_plugin_configuration_absolute_path_values_mock.return_value = ['/mock/test-dir']

        # run test
        actual_test_report_dir = maven_test_reporting_mixin._attempt_get_test_report_directory(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            default='/mock/default',
            require_phase_execution_config=False
        )

        # verify results
        get_plugin_configuration_absolute_path_values_mock.assert_called_once_with(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            work_dir_path='/mock/work-dir-path',
            pom_file='mock-pom.xml',
            profiles=[],
            phases_and_goals=[],
            require_phase_execution_config=False
        )

        self.assertEqual(actual_test_report_dir, '/mock/test-dir')

    def test_two_found_results(self, get_plugin_configuration_absolute_path_values_mock):
        # create object to test against
        maven_test_reporting_mixin = self.__get_maven_test_reporting_mixin()

        # setup mocks
        get_plugin_configuration_absolute_path_values_mock.return_value = [
            '/mock/test-dir1',
            '/mock/test-dir2'
        ]

        # run test
        actual_test_report_dir = maven_test_reporting_mixin._attempt_get_test_report_directory(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            default='/mock/default',
            require_phase_execution_config=False
        )

        # verify results
        get_plugin_configuration_absolute_path_values_mock.assert_called_once_with(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            work_dir_path='/mock/work-dir-path',
            pom_file='mock-pom.xml',
            profiles=[],
            phases_and_goals=[],
            require_phase_execution_config=False
        )

        self.assertEqual(actual_test_report_dir, '/mock/test-dir1')

    def test_found_plugin_but_no_config_use_default(self, get_plugin_configuration_absolute_path_values_mock):
        # create object to test against
        maven_test_reporting_mixin = self.__get_maven_test_reporting_mixin()

        # setup mocks
        get_plugin_configuration_absolute_path_values_mock.return_value = []

        # run test
        actual_test_report_dir = maven_test_reporting_mixin._attempt_get_test_report_directory(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            default='/mock/default',
            require_phase_execution_config=False
        )

        # verify results
        get_plugin_configuration_absolute_path_values_mock.assert_called_once_with(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            work_dir_path='/mock/work-dir-path',
            pom_file='mock-pom.xml',
            profiles=[],
            phases_and_goals=[],
            require_phase_execution_config=False
        )

        self.assertEqual(actual_test_report_dir, '/mock/default')

    def test_plugin_not_found(self, get_plugin_configuration_absolute_path_values_mock):
        # create object to test against
        maven_test_reporting_mixin = self.__get_maven_test_reporting_mixin()

        # setup mocks
        get_plugin_configuration_absolute_path_values_mock.side_effect = RuntimeError(
            'mock could not find plugin error'
        )

        # run test
        with self.assertRaisesRegex(
            StepRunnerException,
            r"Error getting configuration \(mock-reports-dir-config-key\) from maven"
            r" plugin \(mock-maven-test-plugin\):"
            r" mock could not find plugin error"
        ):
            maven_test_reporting_mixin._attempt_get_test_report_directory(
                plugin_name='mock-maven-test-plugin',
                configuration_key='mock-reports-dir-config-key',
                default='/mock/default',
                require_phase_execution_config=False
            )

        # verify results
        get_plugin_configuration_absolute_path_values_mock.assert_called_once_with(
            plugin_name='mock-maven-test-plugin',
            configuration_key='mock-reports-dir-config-key',
            work_dir_path='/mock/work-dir-path',
            pom_file='mock-pom.xml',
            profiles=[],
            phases_and_goals=[],
            require_phase_execution_config=False
        )

@patch("ploigos_step_runner.step_implementers.shared.maven_test_reporting_mixin.aggregate_xml_element_attribute_values")
class TestMavenTestReportingMixin__gather_evidence_from_test_report_directory_testsuite_elements(
    BaseStepImplementerTestCase
):
    def test_found_all_attributes(self, aggregate_xml_element_attribute_values_mock):
        with TempDirectory() as test_dir:
            # setup test
            actual_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            test_report_dir = os.path.join(
                test_dir.path,
                'mock-test-results'
            )

            # setup mocks
            os.mkdir(test_report_dir)
            aggregate_xml_element_attribute_values_mock.return_value = {
                "time": 1.42,
                "tests": 42,
                "errors": 3,
                "skipped": 2,
                "failures": 1
            }

            # run test
            MavenTestReportingMixin._gather_evidence_from_test_report_directory_testsuite_elements(
                step_result=actual_step_result,
                test_report_dir=test_report_dir
            )

            # verify results
            expected_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            expected_step_result.add_evidence(name='time', value=1.42)
            expected_step_result.add_evidence(name='tests', value=42)
            expected_step_result.add_evidence(name='errors', value=3)
            expected_step_result.add_evidence(name='skipped', value=2)
            expected_step_result.add_evidence(name='failures', value=1)
            self.assertEqual(actual_step_result, expected_step_result)

    def test_found_dir_found_no_attributes(self, aggregate_xml_element_attribute_values_mock):
        with TempDirectory() as test_dir:
            # setup test
            actual_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            test_report_dir = os.path.join(
                test_dir.path,
                'mock-test-results'
            )

            # setup mocks
            os.mkdir(test_report_dir)
            aggregate_xml_element_attribute_values_mock.return_value = {
                "time": 1.42,
                "tests": 42
            }

            # run test
            MavenTestReportingMixin._gather_evidence_from_test_report_directory_testsuite_elements(
                step_result=actual_step_result,
                test_report_dir=test_report_dir
            )

            # verify results
            expected_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            expected_step_result.add_evidence(name='time', value=1.42)
            expected_step_result.add_evidence(name='tests', value=42)
            test_report_evidence_element = 'testsuite'
            not_found_attribs = ["errors", "skipped", "failures"]
            expected_step_result.message += "\nWARNING: could not find expected evidence" \
                f" attributes ({not_found_attribs}) on xml element" \
                f" ({test_report_evidence_element}) in test report" \
                f" directory ({test_report_dir})."
            self.assertEqual(actual_step_result, expected_step_result)

    def test_found_dir_found_some_attributes(self, aggregate_xml_element_attribute_values_mock):
        with TempDirectory() as test_dir:
            # setup test
            actual_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            test_report_dir = os.path.join(
                test_dir.path,
                'mock-test-results'
            )

            # setup mocks
            os.mkdir(test_report_dir)
            aggregate_xml_element_attribute_values_mock.return_value = {}

            # run test
            MavenTestReportingMixin._gather_evidence_from_test_report_directory_testsuite_elements(
                step_result=actual_step_result,
                test_report_dir=test_report_dir
            )

            # verify results
            expected_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            test_report_evidence_element = 'testsuite'
            not_found_attribs = ["time", "tests", "errors", "skipped", "failures"]
            expected_step_result.message += "\nWARNING: could not find expected evidence" \
                f" attributes ({not_found_attribs}) on xml element" \
                f" ({test_report_evidence_element}) in test report" \
                f" directory ({test_report_dir})."
            self.assertEqual(actual_step_result, expected_step_result)

    def test_test_report_dir_does_not_exist(self, aggregate_xml_element_attribute_values_mock):
        with TempDirectory() as test_dir:
            # setup test
            actual_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            test_report_dir = os.path.join(
                test_dir.path,
                'mock-test-results'
            )

            # run test
            MavenTestReportingMixin._gather_evidence_from_test_report_directory_testsuite_elements(
                step_result=actual_step_result,
                test_report_dir=test_report_dir
            )

            # verify results
            expected_step_result = StepResult(
                step_name='mock-maven-test-step',
                sub_step_name='mock-maven-test-sub-step',
                sub_step_implementer_name='MockMavenTestReportingMixinStepImplementer'
            )
            expected_step_result.message += f"\nWARNING: test report directory ({test_report_dir})" \
                " does not exist to gather evidence from"
            self.assertEqual(actual_step_result, expected_step_result)



"""Results for Ploigos workflow.
"""

from ploigos_step_runner.results.step_result import StepResult
from ploigos_step_runner.results.step_result_artifact import StepResultArtifact
from ploigos_step_runner.results.step_result_evidence import StepResultEvidence
from ploigos_step_runner.results.workflow_result import WorkflowResult
